# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""API related to node creation in topology"""

import logging
from multiprocessing.dummy import Process
import time

from nest import engine
from nest.topology.interface import BaseInterface, Interface
from nest.engine.t_shark import capture_packets
from nest.engine.util import is_dependency_installed
from nest.topology_map import TopologyMap
from nest import config
from nest.network_utilities import ipv6_dad_check
import nest.global_variables as g_var
from nest.input_validator import input_validator
from .address import Address
from .id_generator import IdGen

logger = logging.getLogger(__name__)

# pylint: disable=too-many-public-methods
class Node:
    """
    Abstraction for a network namespace.

    Attributes
    ----------
    name: str
        User given name for the node.
    """

    @input_validator
    def __init__(self, name: str):
        """
        Create a node with given `name`.

        An unique `id` is assigned to this node which is used by
        `engine` module to create the network namespace.
        This ensures that there is no naming conflict between any two
        nodes.

        Parameters
        ----------
        name: str
            The name of the node to be created
        """
        if name == "":
            raise ValueError("Node name can't be an empty string")
        if config.get_value("assign_random_names") is False and len(name) > 3:
            # We chose 3 because: 'ifb-ns1-ns2-20' is a potential IFB interface name
            # and it's already 14 character long. Note that here node names
            # are 'ns1' and 'ns2'. The `ip` utility won't accept interface names
            # longer than 15 characters
            logger.warning(
                "%s is longer than 3 characters. It's safer to use "
                "node names with atmost 3 characters with the current config.",
                name,
            )

        self._name = name
        self._id = IdGen.get_id(name)
        self.is_mptcp = True
        self.mptcp_monitor_required = False

        # TODO: This list maynot be accurate. Better to refer TopologyMap
        # for list of interfaces in a Node.
        self._interfaces = []

        # Global variable disables when any new node is created
        # to ensure DAD check (if applicable)
        g_var.IS_DAD_CHECKED = False
        self._tshark_processes = []

        engine.create_ns(self.id)
        engine.set_interface_mode(self.id, "lo", "up")
        TopologyMap.add_node(self.id, self)
        TopologyMap.add_host(self)

    def __enter__(self):
        """
        Enter the context of this `Node`.
        For eg., all commands in this context will only be able to see
        interfaces inside this `Node`
        """
        # Go to network namespace corresponding to `Node`
        engine.set_ns(self.id)

    def __exit__(self, *args):
        """
        Exit the context of this `Node`
        """
        # Switch back to default network namespace
        engine.set_ns(None)

    def __del__(self):
        """
        Destructor for node objects
        """
        for process in self._tshark_processes:
            process.join()

    @input_validator
    def set_mptcp_parameters(self, max_subflows: int, max_add_addr_accepted: int):
        """
        Set MPTCP node parameters

        Parameters
        ----------
        max_subflows: int
        max_add_addr_accepted: int
        """
        engine.set_mptcp_node_parameters(max_subflows, max_add_addr_accepted, self.id)

    @input_validator
    def add_mptcp_monitor(self):
        """
        Run MPTCP Monitor on node
        """
        self.mptcp_monitor_required = True

    @input_validator
    def enable_mptcp(self):
        """
        Enable MPTCP sysctl on Node
        """
        self.is_mptcp = True
        engine.enable_mptcp_node(self.id)

    @input_validator
    def disable_mptcp(self):
        """
        Disable MPTCP sysctl on Node
        """
        self.is_mptcp = False
        engine.disable_mptcp_node(self.id)

    def get_mptcp_endpoints(self):
        """
        Gets the MPTCP endpoints of all interfaces on the Node
        """
        return engine.get_mptcp_endpoints(self._id)

    @input_validator
    def add_route(
        self,
        dest_addr: Address,
        via_interface: BaseInterface,
        next_hop_addr: Address = None,
    ):
        """
        Add a route to the routing table of `Node`.

        Parameters
        ----------
        dest_addr: Address/str
            Destination IP address of node to route to. 'DEFAULT' is
            for all addresses
        via_interface: BaseInterface
            `Interface` in `Node` to route via
        next_hop_addr: Address/str, optional
            IP address of next hop Node (or router), by default ''
        """
        if next_hop_addr is None:
            if isinstance(via_interface, Interface) is False:
                raise ValueError(
                    "Unable to automatically determine the next_hop_addr. Please"
                    " provide the parameter explicitly."
                )

            via_interface_addresses = via_interface.pair.get_address(True, True, True)
            if len(via_interface_addresses) > 1:
                raise ValueError(
                    "Please provide the next_hop_addr parameter, since the other"
                    " end of the interface has multiple addresses assigned to it."
                )
            next_hop_addr = via_interface_addresses[0]

        dest_addr_str = ""
        if dest_addr.is_subnet():
            dest_addr_str = dest_addr.get_addr()
        else:
            dest_addr_str = dest_addr.get_addr(with_subnet=False)

        engine.add_route(
            self.id,
            dest_addr_str,
            next_hop_addr.get_addr(with_subnet=False),
            via_interface.id,
        )

    def print_routes_to_file(self, *args):
        """
        Get the routes from the routing table of a Node.

        Parameters
        *args: Interface object
        """

        engine.get_route(self.id, *args)

    @input_validator
    def get_interface(self, node: "Node", connection_number: int = 1):
        """
        Get the interface in `self` connected to `node`.
        By default, this returns a veth interface.

        `connection_number` is an optional argument used when there
        multiple connections between `self` and `node`.

        Parameters
        ----------
        node: Node
            The other end point of the `veth` interface
        connection_number: int
            If there are multiple connections, then this argument
            uniquely identifies the connection

        Returns
        -------
        Interface
            Return the interface between `self` and `node`. Use
            `connection_number` to uniquly identify an interface if
            there are multiple connections.
            If no interface is found, then return None.
        """
        logger.warning(
            "Node.get_interface method will be deprecated soon. Please avoid using this method."
        )

        if connection_number <= 0:
            raise ValueError("connection_number should be greater than 0")

        for interface in self.interfaces:
            if hasattr(interface.pair, "node"):  # True if interface is a `veth`
                pair_node = interface.pair.node
                if node == pair_node:
                    connection_number -= 1

                if connection_number == 0:
                    return interface

        return None

    @input_validator
    def __is_multihomed__(self):
        """
        Is Multihomed?

        This functions returns whether the node has multiple interfaces

        Paramters
        ---------
        self
            Current Node
        """
        return len(self.interfaces) > 1

    @input_validator
    def __is_multiaddressed__(self):
        """
        Is Multiaddressed?

        This functions returns whether the node has multiple gateways
        """
        if not self.__is_multihomed__():
            return False

        differing_subnet_found = False
        subnets_found = {}
        for interface in self.interfaces:
            if interface.get_address().get_subnet() in subnets_found:
                continue

            if len(subnets_found) == 0:
                subnets_found[interface.get_address().get_subnet()] = 1
            else:
                differing_subnet_found = True
                break
        return differing_subnet_found

    @input_validator
    def is_mptcp_supported(self):
        """
        Is MPTCP Supported?

        An MPTCP connection can be setup only if either of the TCP hosts are
        multihomed and multiaddressed. This function returns whether this
        condition is satisfied by the current node (self).
        """
        return self.__is_multiaddressed__() and self.__is_multihomed__()

    @input_validator
    def add_route_mpls_push(
        self, dest_addr: Address, next_hop_addr: Address, label: int
    ):
        """
        Add a route to the routing table of `Node`.

        Parameters
        ----------
        dest_addr: Address/str
            Destination ip address of node to route to. 'DEFAULT' is
            for all addresses
        next_hop_addr: Address/str, optional
            IP address of next hop Node (or router), by default ''
        label: the MPLS label pushed onto the packet
        """
        engine.add_mpls_route_push(
            self.id,
            dest_addr.get_addr(),
            next_hop_addr.get_addr(with_subnet=False),
            label,
        )

    @input_validator
    def add_route_mpls_switch(
        self, incoming_label: int, next_hop_addr: Address, outgoing_label: int
    ):
        """

        Parameters
        ----------
        incoming_label: The label that the packet carries when entering the interface
        next_hop_addr: IP address of the next hop
        outgoing_label: The label that is pushed onto the packet

        """
        engine.add_mpls_route_switch(
            self.id,
            incoming_label,
            next_hop_addr.get_addr(with_subnet=False),
            outgoing_label,
        )

    @input_validator
    def add_route_mpls_pop(self, incoming_label: int, next_hop_addr: Address):
        """

        Parameters
        ----------
        incoming_label: THe label that the packet carries when entering the interface
        next_hop_addr: IP address of the next hop

        """
        engine.add_mpls_route_pop(
            self.id, incoming_label, next_hop_addr.get_addr(with_subnet=False)
        )

    def _add_interface(self, interface: BaseInterface):
        """
        Add `interface` to `Node`

        Parameters
        ----------
        interface: BaseInterface
            `Interface` to be added to `Node`
        """
        self._interfaces.append(interface)
        interface.node_id = self.id
        engine.add_int_to_ns(self.id, interface.id)

    def configure_tcp_param(self, param, value):
        """
        Configure TCP parameters of `Node` available at
        /proc/sys/net/ipv4/tcp_*.

        Example: 'window_scaling', 'wmem', 'ecn', etc.

        If TCP Parameter `param` is valid, then new `value` is set
        for this `param`.

        Parameters
        ----------
        param: str
            TCP parameter to be configured
        value: str
            New value of TCP parameter `param`
        """
        engine.configure_kernel_param(self.id, "net.ipv4.tcp_", param, value)

    def configure_udp_param(self, param, value):
        """
        Configure UDP parameters of `Node` available at
        /proc/sys/net/ipv4/udp_*.

        Example: 'early_demux', 'l3mdev_accept', 'rmem_min', 'wmem_min'

        If UDP Parameter `param` is valid, then new `value` is set
        for this `param`.

        Parameters
        ----------
        param: str
            TCP parameter to be configured
        value: str
            New value of TCP parameter `param`
        """
        engine.configure_kernel_param(self.id, "net.ipv4.udp_", param, value)

    def read_tcp_param(self, param):
        """
        Read TCP parameters of `Node` available at
        `/proc/sys/net/ipv4/tcp_*`.

        Example: 'window_scaling', 'wmem', 'ecn', etc.

        Parameters
        ----------
        param: str
            TCP parameter to be read

        Returns
        -------
        str
            If TCP Parameter `param` is valid, then corresponding value
            is returned.
        """
        return engine.read_kernel_param(self.id, "net.ipv4.tcp_", param)

    def read_udp_param(self, param):
        """
        Read UDP parameters of `Node`available at
        `/proc/sys/net/ipv4/udp_*`.

        Example: 'early_demux', 'l3mdev_accept', 'rmem_min', 'wmem_min'

        Parameters
        ----------
        param: str
            UDP parameter to be read

        Returns
        -------
        str
            If UDP Parameter `param` is valid, then corresponding value
            is returned.
        """
        return engine.read_kernel_param(self.id, "net.ipv4.udp_", param)

    def get_arp_table(self):
        """
        Retrieve the ARP table for the current `Node`

        Returns
        -------
        bool
            Returns `True` if the ARP table is successfully retrieved from the `Node`.
            Else `False`.
        """
        print("ARP table for " + self.name)
        status = engine.get_arp_table(
            self.id,
        )
        return status

    @input_validator
    def delete_arp_entry(self, ip_address: Address):
        """
        Delete an ARP table entry for a specified IP address from the current `Node`

        Parameters
        ----------
        ip_address : Address
            The IP address of the ARP table entry to be deleted.

        Returns
        -------
        bool
            Returns `True` if the ARP table entry for the specified IP address\
            is successfully deleted from the `Node`.
            Else `False`
        """
        logger.info(
            "Deleting ARP table entry for IP address: %s",
            ip_address.get_addr(with_subnet=False),
        )
        status = engine.delete_arp_entry(
            self.id, ip_address.get_addr(with_subnet=False)
        )
        return status

    def flush_arp_table(self):
        """
        Delete all ARP table entries for the current Node.

        Returns
        -------
        bool
            Returns `True` if all entries were deleted successfully.
            Else `False`
        """
        logger.info("Flushing ARP table for: %s", self.name)
        status = engine.flush_arp_table(self.id)
        return status

    @input_validator
    def get_arp_entry(self, ip_address: Address):
        """
        Retrieve the ARP table entry for a specified IP address from the current Node.

        Parameters
        ----------
        ip_address : Address
            The IP address for which to retrieve the ARP table entry.

        Returns
        -------
        bool
            Returns `True` if the entry for the specified IP address was received successfully.
            Else `False`
        """
        status = engine.get_arp_entry(self.id, ip_address.get_addr(with_subnet=False))
        return status

    def set_arp_entry(self, ip_address: Address, mac_address: str):
        """
        Set the ARP table entry for a specified IP address and MAC address from the given Node.

        Parameters
        ----------
        ip_address : Address
            The IP address for which to retrieve the ARP table entry.
        mac_address : string
            The MAC address for which to set the ARP table entry.

        Returns
        -------
        bool
            Returns `True` if the entry for the specified IP address was received successfully.
            Else `False`

        """
        status = engine.set_arp_entry(
            self.id, ip_address.get_addr(with_subnet=False), mac_address
        )
        logger.info(
            "ARP entry set for IP address: %s and MAC address: %s",
            ip_address.get_addr(with_subnet=False),
            mac_address,
        )
        return status

    @ipv6_dad_check
    @input_validator
    def ping(
        self,
        destination_address: Address,
        preload: int = 1,
        packets: int = 5,
        verbose: int = 2,
    ):
        """
        Ping from current `Node` to destination address
        if there is a route.

        Parameters
        ----------
        destination_address: Address/str
            IP address to ping to
        preload: int (Default is 1)
            Number of packets sent as fast as possible without
            waiting for reply.
        packets: int
            Number of ping packets sent
        verbose: int
            If verbose = '0', no output is printed to stdout.
            If verbose = '1', output if ping succeeded or failed.
            If verbose = '2', output details of each ping packet.

        Returns
        -------
        bool
            `True` if `Node` can successfully ping `destination_address`.
            Else `False`.
        """
        if verbose not in [0, 1, 2]:
            raise ValueError(
                f"Verbose parameter value is {verbose}. It should be 0, 1 or 2."
            )

        if verbose == 2:
            print()
            print(
                f"=== PING from {self.name} to "
                f"{destination_address.get_addr(with_subnet=False)} ==="
            )
            print()

        status = engine.ping(
            self.id,
            destination_address.get_addr(with_subnet=False),
            preload,
            packets,
            destination_address.is_ipv6(),
            verbose == 2,
        )

        if verbose == 1:
            print()
            if status is True:
                print(
                    f"SUCCESS : === PING from {self.name} to "
                    f"{destination_address.get_addr(with_subnet=False)} ==="
                )
            elif status is False:
                print(
                    f"FAILURE: === PING from {self.name} to "
                    f"{destination_address.get_addr(with_subnet=False)} ==="
                )
            print()

        return status

    def enable_ip_forwarding(self, ipv4=True, ipv6=True):
        """
        Enable IP forwarding in `Node`.

        After this method runs, the `Node` can be used as a router.
        """
        if not ipv4 and not ipv6:
            raise Exception(
                "IP Forwarding cannot be false for both IPv4 and IPv6 addresses"
            )

        engine.en_ip_forwarding(self.id, ipv4, ipv6)
        TopologyMap.add_router(self)

    def disable_ip_dad(self):
        """
        Disables Duplicate addresses Detection (DAD) for all interfaces of `Node`.
        """
        for i in range(len(self._interfaces)):
            interface_name = self._interfaces[i]
            interface_name.disable_ip_dad()

    # pylint: disable=too-many-arguments
    @input_validator
    def capture_packets(
        self,
        interface: BaseInterface = None,
        packet_count: int = None,
        timeout: int = None,
        output_file: str = None,
        output_format: str = None,
        **kwargs,
    ):
        """
        Packets are captured from this node

        Parameters
        ----------
        interface: BaseInterface
            If the packets need to be captured from only one specific interface
        """

        install_status = is_dependency_installed("tshark")

        if install_status is False:
            logger.error(
                "The system doesn't have tshark installed\n"
                "Packet capture method requires this tool"
            )

        timestamp = time.strftime("%d-%m-%Y-%H:%M:%S")

        if interface is not None:
            kwargs["-i"] = interface.id

        if packet_count is not None:
            kwargs["-c"] = str(packet_count)

        if timeout is not None:
            kwargs["-a"] = f"duration:{timeout}"

        if output_file is None:
            if interface is None:
                output_file = f"{self.name}_{timestamp}"
            else:
                output_file = f"{self.name}_{interface.name}_{timestamp}"
        kwargs["-w"] = output_file

        if output_format is not None:
            kwargs["-T"] = output_format

        t_shark_process = Process(
            target=capture_packets, args=(self.id,), kwargs=kwargs
        )
        self._tshark_processes.append(t_shark_process)
        t_shark_process.start()

    @property
    def id(self):
        """
        str: Value used by `engine` to create the emulated `Node` entity
        """
        return self._id

    @property
    def interfaces(self):
        """list(Interface): List of interfaces in this node"""
        return self._interfaces

    @property
    def name(self):
        """str: User given name for `Node`"""
        return self._name

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.name!r})"



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from routing sub-package"""
import unittest
from os.path import isfile
from nest import config
from nest.topology_map import TopologyMap
from nest.topology import Node, connect
from nest.routing.routing_helper import RoutingHelper
from nest.clean_up import delete_namespaces

# pylint: disable=missing-docstring
class TestFrr(unittest.TestCase):

    # pylint: disable=invalid-name
    def setUp(self):

        self.assertTrue(isfile("/usr/lib/frr/zebra"), "Frrouting is not installed")

        self.n0 = Node("n0")
        self.n1 = Node("n1")
        self.r0 = Node("r0")
        self.r1 = Node("r1")
        self.r0.enable_ip_forwarding()
        self.r1.enable_ip_forwarding()

        ### Create interfaces and connect nodes and routers ###

        (eth_p1r1, eth_r1p1) = connect(self.n0, self.r0, "eth-n1r1-0", "eth-r1n1-0")
        (eth_r1r2, eth_r2r1) = connect(self.r0, self.r1, "eth-r1r2-0", "eth-r2r1-0")
        (eth_r2p2, eth_p2r2) = connect(self.r1, self.n1, "eth-r2n2-0", "eth-n2r2-0")

        ### Assign addresses to interfaces ###

        eth_p1r1.set_address(["10.0.1.1/24", "10::1:1/122"])
        eth_r1p1.set_address(["10.0.1.2/24", "10::1:2/122"])

        eth_r1r2.set_address(["10.0.2.2/24", "10::2:2/122"])
        eth_r2r1.set_address(["10.0.2.3/24", "10::2:3/122"])

        eth_r2p2.set_address(["10.0.3.3/24", "10::3:3/122"])
        eth_p2r2.set_address(["10.0.3.4/24", "10::3:4/122"])

        config.set_value("routing_suite", "frr")  # Use frr

        self.routing_helper_ipv4 = None
        self.routing_helper_ipv6 = None

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()
        if self.routing_helper_ipv4:
            # pylint: disable=protected-access
            self.routing_helper_ipv4._clean_up()
        if self.routing_helper_ipv6:
            # pylint: disable=protected-access
            self.routing_helper_ipv6._clean_up()

    def test_routing_helper(self):

        self.routing_helper_ipv4 = RoutingHelper("rip")
        self.routing_helper_ipv6 = RoutingHelper("rip", ipv6_routing=True)
        self.routing_helper_ipv4.populate_routing_tables()
        self.routing_helper_ipv6.populate_routing_tables()

        status_ipv4 = self.n0.ping("10.0.3.4", verbose=False)
        self.assertTrue(status_ipv4)
        status_ipv6 = self.n0.ping("10::3:4", verbose=False)
        self.assertTrue(status_ipv6)

        status_ipv4 = self.n0.ping("10.0.1.1", verbose=False)
        self.assertTrue(status_ipv4)
        status_ipv6 = self.n1.ping("10::1:1", verbose=False)
        self.assertTrue(status_ipv6)

    def test_ospf(self):
        self.routing_helper_ipv4 = RoutingHelper("ospf")
        self.routing_helper_ipv6 = RoutingHelper("ospf", ipv6_routing=True)
        self.routing_helper_ipv4.populate_routing_tables()
        self.routing_helper_ipv6.populate_routing_tables()

        status_ipv4 = self.n0.ping("10.0.3.4", verbose=False)
        self.assertTrue(status_ipv4)
        status_ipv6 = self.n0.ping("10::3:4", verbose=False)
        self.assertTrue(status_ipv6)

        status_ipv4 = self.n0.ping("10.0.1.1", verbose=False)
        self.assertTrue(status_ipv4)
        status_ipv6 = self.n1.ping("10::1:1", verbose=False)
        self.assertTrue(status_ipv6)

    def test_isis(self):
        self.routing_helper_ipv4 = RoutingHelper("isis")
        self.routing_helper_ipv6 = RoutingHelper("isis", ipv6_routing=True)
        self.routing_helper_ipv4.populate_routing_tables()
        self.routing_helper_ipv6.populate_routing_tables()

        status_ipv4 = self.n0.ping("10.0.3.4", verbose=False)
        self.assertTrue(status_ipv4)
        status_ipv6 = self.n0.ping("10::3:4", verbose=False)
        self.assertTrue(status_ipv6)

        status_ipv4 = self.n0.ping("10.0.1.1", verbose=False)
        self.assertTrue(status_ipv4)
        status_ipv6 = self.n1.ping("10::1:1", verbose=False)
        self.assertTrue(status_ipv6)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal
"""Test APIs from topology packet operations"""

import unittest
from contextlib import redirect_stdout
import io
import os
from os.path import exists
from nest.topology import Node, connect
from nest.clean_up import delete_namespaces
from nest.topology_map import TopologyMap


# pylint: disable=invalid-name
# pylint: disable=missing-docstring


class TestTopologyPacketOps(unittest.TestCase):
    def setUp(self):
        self.n0 = Node("n0")
        self.n1 = Node("n1")

        (self.n0_n1, self.n1_n0) = connect(self.n0, self.n1)

        self.n0_n1.set_address("10.0.0.1/24")
        self.n1_n0.set_address("10.0.0.2/24")

    def parse_ping_output(self):
        # TODO: Optimize the below function using regex

        f = io.StringIO()
        checker = False
        with redirect_stdout(f):
            self.n0.ping("10.0.0.2", packets=50)
        out = f.getvalue()
        pointer = out.find(
            "packet loss"
        )  # pointer to the initial character of string "packet loss".
        pointer = pointer - 3  # it will point to the last digit of packet loss %.
        num = 0
        while (
            out[pointer] != " "
        ):  # it is tracing from last digit to the first digit to get actual % loss number.
            num = int(out[pointer]) * 10 + num
            pointer = pointer - 1
        if num > 0:
            checker = True
        self.assertTrue(checker, "packets are not being corrupted")

    def test_packet_loss_state(self):
        self.n0_n1.set_packet_loss_state("20%", "4%", "5%", "4%", "3%", True)
        status = self.n0.ping("10.0.0.2")

        self.assertTrue(status)

    def test_packet_loss_gemodel(self):
        self.n0_n1.set_packet_loss_gemodel("10%", "4%", "5%", "4%", True)
        status = self.n0.ping("10.0.0.2")

        self.assertTrue(status)

    # Add rate in percent to get packet duplicated.
    def test_packet_duplication(self):
        # Add rate in percent to get packet duplicated.
        self.n0_n1.set_packet_duplication("50%")

        f = io.StringIO()
        checker = False
        with redirect_stdout(f):
            self.n0.ping("10.0.0.2", packets=15)
        out = f.getvalue()
        if "duplicates" in out:
            checker = True
        self.assertTrue(checker, "Packets are not being duplicated")

    def test_packet_corruption(self):
        self.n0_n1.set_packet_corruption("75%", "50%")
        self.parse_ping_output()

    def test_packet_reordering(self):
        self.n0_n1.set_attributes("10mbit", "10ms")
        self.n1_n0.set_attributes("10mbit", "5ms")

        self.n0_n1.set_packet_reordering("10ms", "25%", gap=5)

        self.n0.ping("10.0.0.2", packets=1)  # build ARP table before preload is used.
        status = self.n0.ping("10.0.0.2", preload=10, packets=10)

        self.assertTrue(status)

    def test_delay_distribution(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        n0_n1.set_attributes("10mbit", "10ms")
        n1_n0.set_attributes("10mbit", "5ms")

        n0_n1.set_delay_distribution("100ms", "10ms", "normal")

        status = self.n0.ping("10.0.0.2", packets=10)

        self.assertTrue(status)

    def test_packet_capture(self):
        self.n0_n1.set_attributes("10mbit", "10ms")
        current_dir = os.getcwd()
        initial_status = exists(f"{current_dir}/packet_capture/packet_capture.pcap")
        self.n1.capture_packets(
            interface=self.n1_n0, packet_count=10, output_file="packet_capture.pcap"
        )
        self.n0.ping("10.0.0.2", packets=10)

        status = exists(f"{current_dir}/packet_capture/packet_capture.pcap")
        self.assertTrue(status)
        if initial_status is False and status is True:
            os.remove(f"{current_dir}/packet_capture/packet_capture.pcap")

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Class to handles OSPF related functionalities"""

import random
from functools import partial
from nest.engine.dynamic_routing import run_ospfd
from nest.routing.route_daemons import RoutingDaemonBase
from nest import config

# pylint: disable=line-too-long


class Ospf(RoutingDaemonBase):
    """
    Handles OSPF related functionalities.
    """

    def __init__(self, router_ns_id, ipv6_routing, interfaces, conf_dir, **kwargs):
        if config.get_value("routing_suite") == "bird":
            super().__init__(
                router_ns_id, ipv6_routing, interfaces, "bird", conf_dir, **kwargs
            )
        else:
            super().__init__(
                router_ns_id, ipv6_routing, interfaces, "ospfd", conf_dir, **kwargs
            )

    # pylint: disable=too-many-branches
    def create_basic_config(self):
        """
        Creates a file with basic configuration for OSPF.
        Use base `add_to_config` directly for more complex configurations
        """
        if config.get_value("routing_suite") == "bird":
            if not self.ipv6_routing:
                self.conf.write(
                    "protocol kernel{\n\tipv4{\n\t\texport all;\n\t};\n\tpersist;\n\tlearn;\n}\n"
                )
                self.conf.write("protocol static{\n\tipv4;\n}\n")
                self.conf.write("protocol device{\n}\n")
                self.conf.write(
                    'protocol ospf v2{\n\tarea 0{\n\t\tinterface "*"{\n\t\t\ttype pointopoint;\n\t\t};\n\t};\n}\n'
                )
            else:
                self.conf.write(
                    "protocol kernel{\n\tipv6{\n\t\texport all;\n\t};\n\tpersist;\n\tlearn;\n}\n"
                )
                self.conf.write("protocol static{\n\tipv6;\n}\n")
                self.conf.write("protocol device{\n}\n")
                self.conf.write(
                    'protocol ospf v3{\n\tarea 0{\n\t\tinterface "*"{\n\t\t\ttype pointopoint;\n\t\t};\n\t};\n}\n'
                )
        else:
            if self.ipv6_routing:
                for interface in self.interfaces:
                    self.add_to_config(f"interface {interface.id}")
                    # send hello packets every 1 second for faster convergence
                    self.add_to_config("ipv6 ospf6 hello-interval 1")

                self.add_to_config("router ospf6")

                # Generates random router-id in A.B.C.D format
                router_id = ".".join(
                    map(str, (random.randint(0, 255) for _ in range(0, 4)))
                )
                # for quagga
                if config.get_value("routing_suite") == "quagga":
                    self.add_to_config(f"router-id {router_id}")
                # for frr
                else:
                    self.add_to_config(f"ospf6 router-id {router_id}")
                for interface in self.interfaces:
                    self.add_to_config(f" interface {interface.id} area 0.0.0.0")
            else:
                for interface in self.interfaces:
                    self.add_to_config(f"interface {interface.id}")
                    # send hello packets every 1 second for faster convergence
                    self.add_to_config("ip ospf hello-interval 1")

                self.add_to_config("router ospf")
                self.add_to_config(
                    f"ospf router-id {self.interfaces[0].get_address(not self.ipv6_routing, self.ipv6_routing, True)[0].get_addr(with_subnet=False)}"
                )
                for interface in self.interfaces:
                    self.add_to_config(
                        f" network {interface.get_address(not self.ipv6_routing, self.ipv6_routing, True)[0].get_subnet()} area 0.0.0.0"
                    )

            if self.log_file is not None:
                self.add_to_config(f"log file {self.log_file}")

            self.create_config()

    def run(self):
        """
        Runs the ospfd command
        """
        super().run(
            engine_func=partial(
                run_ospfd,
                self.router_ns_id,
                self.conf_file,
                self.pid_file,
                self.ipv6_routing,
                log_file=self.log_file,
                socket_file=self.socket_file,
            )
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from topology sub-package"""

import os
import json
import shutil
import unittest
from nest.topology import Node, connect
from nest.experiment import Experiment, Flow
from nest.clean_up import delete_namespaces
from nest.topology_map import TopologyMap


# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=too-many-locals
class TestMPTCP(unittest.TestCase):
    def setUp(self):
        # Create two MPTCP enabled hosts `h1` and `h2`
        h1 = Node("h1")
        h2 = Node("h2")

        # Connect `h1` to `h2`
        # `eth1a` and `eth2a` are the interfaces at `h1` and `h2`, respectively, for 1st connection.
        # `eth1b` and `eth2b` are the interfaces at `h1` and `h2`, respectively, for 2nd connection.
        (eth1a, eth2a) = connect(h1, h2)
        (eth1b, eth2b) = connect(h1, h2)

        # Assign IPv4 addresses to all the interfaces.
        eth1a.set_address("10.10.0.1/24")
        eth2a.set_address("10.10.0.2/24")
        eth1b.set_address("192.168.0.1/24")
        eth2b.set_address("192.168.0.2/24")

        # Set the link attributes: `h1` <--> `h2`
        eth1a.set_attributes("10mbit", "5ms")
        eth2a.set_attributes("10mbit", "5ms")
        eth1b.set_attributes("5mbit", "10ms")
        eth2b.set_attributes("5mbit", "10ms")

        h1.set_mptcp_parameters(1, 1)
        h2.set_mptcp_parameters(1, 1)

        eth1b.enable_mptcp_endpoint(["subflow"])
        eth2b.enable_mptcp_endpoint(["signal"])

        self.exp = Experiment("test-experiment-mptcp")
        flow = Flow(
            h1, h2, eth2a.get_address(), 0, 20, 1, source_address=eth1a.get_address()
        )
        self.exp.add_mptcp_flow(flow)
        self.exp.run()

    def test_mptcp(self):
        for directory in sorted(os.listdir("./")):
            if self.exp.name in directory:
                # pylint: disable=attribute-defined-outside-init
                self.experiment_directory = os.path.abspath(directory)

        self.assertIn("netperf.json", os.listdir(self.experiment_directory))

        with open(os.path.join(self.experiment_directory, "netperf.json"), "r") as f:
            experiment_data = json.load(f)["h1"][0]

            rate_counter = 0
            rate_sum = 0.0
            for obj in list(experiment_data.values())[0][11:]:
                rate_sum += float(obj["sending_rate"])
                rate_counter += 1

            self.assertTrue(
                rate_sum / rate_counter > 10, "MPTCP doesn't seem to be working!"
            )

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()
        shutil.rmtree(self.experiment_directory)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""Check if tool is installed and tool options are valid"""

from nest.engine.util import is_dependency_installed, is_package_installed
from nest.input_validator import input_validator


# pylint: disable=too-few-public-methods, invalid-name, missing-class-docstring
class Tools:
    """
    Check that the tool is available.
    """

    @input_validator
    def __init__(self, tool_name: str):
        """
        Parameters
        ----------
        tool_name: str
            name of tool

        """

        self.tool_name = tool_name

        if (self.tool_name == "coap") and not is_package_installed("aiocoap"):
            raise ValueError("coap is not installed.")

        if not is_dependency_installed(self.tool_name):
            raise ValueError(f"{self.tool_name} is either invalid or not installed.")

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.tool_name!r})"


class Options:
    # global variable
    user_options = {}

    def __init__(self):
        self.selected_options = {}


# pylint: disable= consider-using-dict-items, redefined-builtin
class Iperf3(Options):
    """
    Validate whether Selected options and there value are correct

    Returns:
        user_selected_options: dict
            A dictionary with correct options
    """

    # pylint: disable= too-many-arguments, unused-argument
    @staticmethod
    @input_validator
    def server_option(
        s_verbose: bool = False,
        s_interval: float = 0.2,
        s_format: str = None,
        s_logfile: str = None,
        s_forceflush: bool = False,
        s_timestamps: str = None,
        port_no: int = None,
        one_off: bool = True,
        bitrate: str = None,
        daemon: bool = False,
        bind: str = None,
    ):
        """
        select options to configure iperf3 server

        Returns:
            selected_option: dict
                valid selected options for iperf3 server configuration
        """
        Iperf3.user_options = {}
        for option, val in locals().items():
            if val:
                Iperf3.user_options.update({option: val})
        return Iperf3.user_options

    @staticmethod
    @input_validator
    def client_option(
        verbose: bool = False,
        interval: float = 0.2,
        format: str = None,
        logfile: str = None,
        forceflush: bool = False,
        timestamps: str = False,
        cport: int = None,
        target_bw: str = None,
        bind: str = None,
    ):
        """
        select options to configure iperf3 client

        Returns:
            selected_option: dict
                valid selected options for iperf3 client configuration
        """
        Iperf3.user_options = {}
        for option, val in locals().items():
            if val:
                Iperf3.user_options.update({option: val})
        return Iperf3.user_options

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.user_options!r})"


@input_validator
class Iperf3Options(Options):
    """
    Validate whether Selected options and there value are correct

    Returns:
        user_selected_options: dict
            A dictionary with correct options
    """

    # server only options
    s_verbose: bool
    s_interval: float
    s_format: str
    s_logfile: str
    s_forceflush: bool
    s_timestamps: str
    port_no: int
    one_off: bool
    bitrate: str
    daemon: bool

    # client only options
    verbose: bool
    interval: float
    format: str
    logfile: str
    forceflush: bool
    forceflush: bool
    timestamps: str
    cport: int
    protocol: str
    target_bw: str
    cong_algo: str
    kwargs: dict

    # general options
    bind: str

    # pylint: disable= consider-using-dict-items
    def __init__(self, protocol="tcp", kwargs: dict = None):
        super().__init__()

        default_options = {
            "default_server_options": {
                "s_verbose": False,  # more detailed output as a log file
                "s_interval": 0.2,  # Generate interim results every INTERVAL seconds
                "s_format": None,  # format to report: Kbits, Mbits, Gbits, Tbits
                "s_logfile": None,  # send output to a log file
                "s_forceflush": False,  # force flushing output at every interval
                "s_timestamps": None,  # generate timestamp at the start of each output line
                "port_no": None,  # server port to listen on(#generated randomely)
                "one_off": True,  # handle one client connection then exit
                "bitrate": None,  # server's total bit rate limit
                "daemon": False,  # run server as daemon
                "bind": None,  # bind to the given interface
            },
            "default_client_options": {
                "verbose": False,  # more detailed output as a log file
                "interval": 0.2,  # Generate interim results every INTERVAL seconds
                "format": None,  # format to report: Kbits, Mbits, Gbits, Tbits
                "logfile": None,  # send output to a log file
                "forceflush": False,  # force flushing output at every interval
                "timestamps": None,  # generate timestamp at the start of each output line
                "cport": None,  # bind to a specific client port
                "protocol": None,  # protocol name which is going to be used
                "target_bw": "1mbit",  # target bitrate in bits/sec at client side
                "bind": None,  # bind to the given interface
                # "cong_algo": "cubic",  # set TCP congestion control algorithm
            },
        }

        if protocol == "tcp":
            # Add cong_algo only for tcp protocol
            default_options["default_client_options"][
                "cong_algo"
            ] = "cubic"  # set TCP congestion control algorithm
        elif protocol == "udp":
            # Nothing specific to do as of now
            pass
        else:
            raise ValueError("Invalid protocol string. It should be 'tcp' or 'udp'.")

        # Validate the dictionary passed and set the internal data members
        for _, option in default_options.items():
            for option_key in option:
                if option_key in kwargs:
                    if kwargs[option_key]:
                        self.selected_options.update({option_key: kwargs[option_key]})
                    continue
                if option[option_key]:
                    self.selected_options.update({option_key: option[option_key]})

    def getter(self):
        """
        Get Iperf3 command line options based on user selected options
        """
        return self.selected_options

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.selected_options!r})"



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""API related to creation of bridge in topology"""

import logging

from nest import engine
from .device import Device

logger = logging.getLogger(__name__)


class Bridge(Device):
    """
    Abstraction for a bridge.

    Attributes
    ----------
    id : str
        This value is used by `engine` to create emulated interface
        entity
    node_id : str
        id of the Node to which this Device belongs
    """

    def __init__(self, name, node_id):
        """
        Constructor of Bridge

        Parameters
        ----------
        name : str
            Name of the bridge
        node_id : str
            Id of the node that the bridge belongs to
        """
        engine.create_switch(node_id, node_id)
        engine.set_switch_mode(node_id, "up")
        super().__init__(name, node_id)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""
This module contains methods to kill any running processes in namespaces
and delete all namespaces after the experiment is complete.
"""

import atexit
import logging
import shutil
from pathlib import Path
from nest.network_utilities import ipv6_dad_check
from nest.mpeg_dash_encoder import MpegDashEncoder
from nest import engine
from . import config

from . import engine
from .topology_map import TopologyMap

logger = logging.getLogger(__name__)


# pylint: disable=import-outside-toplevel
# pylint: disable=cyclic-import
@atexit.register
def tcp_modules_clean_up():
    """Clean up the modified TCP modules"""

    from .experiment.experiment import Experiment

    # Remove newly loaded modules
    for cong_algo in Experiment.new_cong_algos:
        engine.remove_tcp_module(cong_algo)
    (Experiment.new_cong_algos).clear()

    # Reset old modules with original params
    for cong_algo, params in (Experiment.old_cong_algos).items():
        engine.set_tcp_params(cong_algo, params, True)
    (Experiment.old_cong_algos).clear()


def kill_processes():
    """
    Kill any running processes in namespaces
    """
    nodes = TopologyMap.get_nodes()

    for ns_id in nodes:
        engine.kill_all_processes(ns_id)


@atexit.register
@ipv6_dad_check
def delete_namespaces():
    """
    Delete all the newly generated namespaces
    """
    nodes = TopologyMap.get_nodes()

    if config.get_value("delete_namespaces_on_termination"):
        for ns_id in nodes:
            engine.delete_ns(ns_id)
        logger.info("Cleaned up environment!")
    else:
        logger.info("Namespaces not deleted")


@atexit.register
def delete_encoded_mpeg_dash_chunks():
    """
    Delete the encoded MPEG-DASH chunks
    """

    if config.get_value("mpeg_dash_delete_encoded_chunks_on_termination"):
        dirs_to_delete = MpegDashEncoder.mpeg_dash_encoded_chunks_path_list
        for directory in dirs_to_delete:
            if Path(directory).exists():
                shutil.rmtree(directory)
                logger.info(
                    "Deleted encoded MPEG-DASH chunks at path : %s !", directory
                )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from static routing"""

import unittest
from nest.topology import Node, connect
from nest.clean_up import delete_namespaces
from nest.topology_map import TopologyMap
from nest.routing.routing_helper import RoutingHelper

# pylint: disable=missing-docstring


class TestStaticRouting(unittest.TestCase):
    # pylint: disable=invalid-name
    def setUp(self):
        # create nodes
        # pylint: disable=anomalous-backslash-in-string
        """
        Topology is as follows:

        n0            n2
          \          /
           \        /
            r0----r1
           /        \
          /          \
        n1            n3
        """

        self.n0 = Node("n0")
        self.n1 = Node("n1")
        self.n2 = Node("n2")
        self.n3 = Node("n3")

        self.r0 = Node("r0")
        self.r1 = Node("r1")
        self.r0.enable_ip_forwarding()
        self.r1.enable_ip_forwarding()

        # populated later when we set the addresses
        # to be used for pairwise ping tests
        self.nodes_addresses = []

        (eth_n0_r0, eth_r0_n0) = connect(self.n0, self.r0, "eth-n0_r0", "eth-r0_n0")
        (eth_n1_r0, eth_r0_n1) = connect(self.n1, self.r0, "eth-n1_r0", "eth-r0_n1")

        (eth_r0_r1, eth_r1_r0) = connect(self.r0, self.r1, "eth-r0_r1", "eth-r1_r0")

        (eth_n2_r1, eth_r1_n2) = connect(self.n2, self.r1, "eth-n2_r1", "eth-r1_n2")
        (eth_n3_r1, eth_r1_n3) = connect(self.n3, self.r1, "eth-n3_r1", "eth-r1_n3")

        ### Assign addresses to interfaces ###
        eth_n0_r0.set_address("10.0.1.1/24")
        eth_r0_n0.set_address("10.0.1.2/24")
        self.nodes_addresses.append((self.n0, "10.0.1.1"))

        eth_n1_r0.set_address("10.0.1.3/24")
        eth_r0_n1.set_address("10.0.1.4/24")
        self.nodes_addresses.append((self.n1, "10.0.1.3"))

        eth_r0_r1.set_address("10.0.2.5/24")
        eth_r1_r0.set_address("10.0.2.6/24")

        eth_r1_n2.set_address("10.0.3.7/24")
        eth_n2_r1.set_address("10.0.3.8/24")
        self.nodes_addresses.append((self.n2, "10.0.3.8"))

        eth_r1_n3.set_address("10.0.3.9/24")
        eth_n3_r1.set_address("10.0.3.10/24")
        self.nodes_addresses.append((self.n3, "10.0.3.10"))

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()

        ### Create interfaces and connect nodes and routers ###

    def test_routing(self):
        # this runs a DFS algorithm and populates routing tables accordingly
        # If the graph is a tree, there is only one path for each pair\
        # for non-trees, this will use a spanning tree found by DFS
        RoutingHelper("static").populate_routing_tables()

        # Ping between each pair of nodes
        for node1, address1 in self.nodes_addresses:
            for _, address2 in self.nodes_addresses:
                if address1 != address2:
                    status = node1.ping(address2, verbose=2)
                    self.assertTrue(status)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
This folders contains implementation for various interfaces
which will be used by NeST users.
"""

from .base_interface import BaseInterface
from .interface import Interface, create_veth_pair



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to Address Helper"""

import logging
from nest.input_validator.input_validator import input_validator
from nest.topology.network import Network
from nest.topology_map import TopologyMap
from .address import Subnet

# pylint: disable=too-few-public-methods

logger = logging.getLogger(__name__)


class AddressHelper:
    """
    Abstraction for automatic address assignment to interfaces of the network.
    """

    @staticmethod
    @input_validator
    def assign_addresses(network: Network = None):
        """
        Assignment of addresses to the interfaces.

        Parameters
        ----------
        network : Network
            Assigning the addresses to each interfaces of all network.
            It will assign addresses to specific network if network object is mentioned.
        """
        if network is None:

            networks = TopologyMap.get_networks()
            for net in networks:
                AddressHelper.__assign_addresses_to_network(net)

            # Warning for the assignment of the addresses to the interfaces,
            # which do not belong to any network
            if TopologyMap.orphan_interfaces != 0:
                logger.warning(
                    "Interfaces not part of some network. The addresses to these orphaned "
                    "interfaces should be manually added.",
                )
        else:
            AddressHelper.__assign_addresses_to_network(network)

    @staticmethod
    def __assign_addresses_to_network(network: Network):
        """
        Assignment of addresses to all interfaces of specific network.

        Parameters
        ----------
        network : Network
            Assigning the addresses to each interfaces of the given network object.
        """
        _interfaces = []
        _interfaces = network.interfaces

        _net_address = Subnet(network.net_address)

        for inter in _interfaces:
            inter.set_address(_net_address.get_next_addr())



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to interfaces in topology"""

import logging

from nest.input_validator import input_validator
from nest.topology.device import VethEnd
from nest import engine
from .base_interface import BaseInterface

logger = logging.getLogger(__name__)


class Interface(BaseInterface):
    """
    Abstraction for an network interface.

    Attributes
    ----------
    name : str
        User given name for the interface
    id : str
        This value is used by `engine` to create emulated interface
        entity
    node : Node
        `Node` which contains this `Interface`
    address : str/Address
        IP address assigned to this interface
    """

    @input_validator
    def __init__(self, interface_name: str):
        """
        Constructor of Interface.

        *Note*: The creation of Interface object does not
        create any devices in the interface.
        [See `create_veth_pair` method]

        Parameters
        ----------
        interface_name : str
            Name of the interface
        """

        super().__init__(interface_name, VethEnd(interface_name, None))

    @property
    def pair(self):
        """
        Get other pair for this interface (assuming veth)
        """
        return self._pair

    @pair.setter
    def pair(self, interface: "Interface"):
        """
        Setter for the other end of the interface that it is connected to

        Parameters
        ----------
        interface : Interface
            The interface to which this interface is connected to
        """
        self._pair = interface

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.name!r})"


def create_veth_pair(interface1_name, interface2_name):
    """
    Handle creation of Veth pairs between
    `interface1_name` and `interface2_name`

    Parameters
    ----------
    interface1_name : str
            Name of one of the interfaces to be connected
    interface2_name : str
            Name of the other interface to be connected
    """

    interface1 = Interface(interface1_name)
    interface2 = Interface(interface2_name)

    interface1.pair = interface2
    interface2.pair = interface1

    # Create the veth
    engine.create_veth(interface1.id, interface2.id)

    return (interface1, interface2)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from topology sub-package"""

import unittest
import subprocess

from nest.topology import Node, connect, Switch, Router, Address
from nest.clean_up import delete_namespaces
from nest.topology_map import TopologyMap
from nest.topology.interface import Interface
from nest import config

# pylint: disable=too-many-public-methods
# pylint: disable=missing-docstring
# pylint: disable=invalid-name
class TestTopology(unittest.TestCase):
    def setUp(self):
        self.n0 = Node("n0")
        self.n1 = Node("n1")
        self.n2 = Node("n2")
        self.n3 = Node("n3")
        self.n4 = Node("n4")
        self.n5 = Node("n5")

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()

    def test_p2p(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        n0_n1.set_address("10.0.0.1/24")
        n1_n0.set_address("10.0.0.2/24")

        status = self.n0.ping("10.0.0.2")

        self.assertTrue(status)

    def test_p2p_ipv6(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        self.n0.disable_ip_dad()
        self.n1.disable_ip_dad()

        n0_n1.set_address("2001:1:1:1443::411/122")
        n1_n0.set_address("2001:1:1:1443::412/122")

        status = self.n0.ping(n1_n0.address)

        self.assertTrue(status)

    def test_prp(self):
        # pylint: disable=invalid-name
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(self.n0, r)
        (r_n1, n1_r) = connect(r, self.n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        self.n0.add_route("DEFAULT", n0_r)
        self.n1.add_route("DEFAULT", n1_r)

        status = self.n0.ping("10.1.2.1")

        self.assertTrue(status)

    def test_prp_ipv6(self):
        # pylint: disable=invalid-name
        r = Node("r")
        r.enable_ip_forwarding(ipv6=True)

        (n0_r, r_n0) = connect(self.n0, r)
        (r_n1, n1_r) = connect(r, self.n1)

        self.n0.disable_ip_dad()
        self.n1.disable_ip_dad()

        n0_r.set_address("2001:0001:0001:1443::0411/122")
        r_n0.set_address("2001:0001:0001:1443::0412/122")
        r_n1.set_address("2001:0001:0001:1444::0412/122")
        n1_r.set_address("2001:0001:0001:1444::0411/122")

        self.n0.add_route("DEFAULT", n0_r)
        self.n1.add_route("DEFAULT", n1_r)

        status = self.n0.ping(n1_r.address)

        self.assertTrue(status)

    def test_p2p_multiple_set_address(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        self.n0.disable_ip_dad()
        self.n1.disable_ip_dad()

        status_1 = self.n0.ping(Address("10.1.1.2/24"))
        status_2 = self.n0.ping(Address("10.1.1.3/24"))
        status_3 = self.n0.ping(Address("2001:1:1:1443::412/122"))

        self.assertFalse(status_1)
        self.assertFalse(status_2)
        self.assertFalse(status_3)

        n0_n1.set_address(["10.1.1.1/24", "2001:1:1:1443::411/122"])
        n1_n0.set_address(["10.1.1.2/24", "10.1.1.3/24", "2001:1:1:1443::412/122"])

        status_4 = self.n0.ping(n1_n0.get_address(ipv4=True, ipv6=False)[0])
        status_5 = self.n0.ping(n1_n0.get_address(ipv4=True, ipv6=False)[1])
        status_6 = self.n0.ping(n1_n0.get_address(ipv4=False, ipv6=True))

        self.assertTrue(status_4)
        self.assertTrue(status_5)
        self.assertTrue(status_6)

    def test_add_addr(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        self.n0.disable_ip_dad()
        self.n1.disable_ip_dad()

        n0_n1.set_address(["10.1.1.1/24", "2001:1:1:1443::411/122"])
        n1_n0.set_address(["10.1.1.2/24", "2001:1:1:1443::412/122"])

        n1_n0.add_address(["10.1.1.3/24", "2001:1:1:1443::413/122"])

        status_1 = self.n0.ping(n1_n0.get_address(ipv4=True, ipv6=False)[0])
        status_2 = self.n0.ping(n1_n0.get_address(ipv4=True, ipv6=False)[1])
        status_3 = self.n0.ping(n1_n0.get_address(ipv4=False, ipv6=True)[0])
        status_4 = self.n0.ping(n1_n0.get_address(ipv4=False, ipv6=True)[1])

        self.assertTrue(status_1)
        self.assertTrue(status_2)
        self.assertTrue(status_3)
        self.assertTrue(status_4)

    def test_get_addr(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        self.n0.disable_ip_dad()
        self.n1.disable_ip_dad()

        n0_n1.set_address(["10.0.0.1/24", "2001:1:1:1443::411/122"])
        n1_n0.set_address(["10.0.0.2/24", "2001:1:1:1443::412/122"])

        addr = n0_n1.get_address()
        self.assertEqual(
            (addr[0].get_addr(), addr[1].get_addr()),
            (
                Address("10.0.0.1/24").get_addr(),
                Address("2001:1:1:1443::411/122").get_addr(),
            ),
            "n0_n1 interface has unexpected addresses",
        )

        addr = n1_n0.get_address()
        self.assertEqual(
            (addr[0].get_addr(), addr[1].get_addr()),
            (
                Address("10.0.0.2/24").get_addr(),
                Address("2001:1:1:1443::412/122").get_addr(),
            ),
            "n1_n0 interface has unexpected address list",
        )

        addr = n0_n1.get_address(True, False, False)
        self.assertEqual(
            addr.get_addr(),
            Address("10.0.0.1/24").get_addr(),
            "n0_n1 interface has unexpected ipv4 address",
        )

        addr = n0_n1.get_address(False, True, False)
        self.assertEqual(
            addr.get_addr(),
            Address("2001:1:1:1443::411/122").get_addr(),
            "n0_n1 interface has unexpected ipv6 address",
        )

        addr = n0_n1.get_address(True, False, True)
        self.assertEqual(
            [addr[0].get_addr()],
            [Address("10.0.0.1/24").get_addr()],
            "n0_n1 interface has unexpected ipv4 address list",
        )

        addr = n0_n1.get_address(False, True, True)
        self.assertEqual(
            [addr[0].get_addr()],
            [Address("2001:1:1:1443::411/122").get_addr()],
            "n0_n1 interface has unexpected ipv6 address list",
        )

        addr = n0_n1.get_address(True, True, True)
        self.assertEqual(
            (addr[0].get_addr(), addr[1].get_addr()),
            (
                Address("10.0.0.1/24").get_addr(),
                Address("2001:1:1:1443::411/122").get_addr(),
            ),
            "n0_n1 interface has unexpected address lists",
        )

        with self.assertRaises(ValueError) as err:
            addr = n0_n1.get_address(False, False, False)
        self.assertEqual(
            str(err.exception), "Either ipv4 or ipv6 argument must be True"
        )

    def test_del_addr(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        n0_n1.set_address(["10.1.1.1/24", "2001:1:1:1443::411/122"])
        n1_n0.set_address(["10.1.1.2/24", "10.1.1.3/24", "2001:1:1:1443::412/122"])

        # Test 1
        status_1 = self.n0.ping(Address("10.1.1.2/24"))

        self.assertTrue(status_1, "Ping failed")

        n1_n0.del_address("10.1.1.2/24")

        status_2 = self.n0.ping(Address("10.1.1.2/24"))

        self.assertFalse(status_2, "Address not deleted")

        addr = n1_n0.get_address(True, False, False)
        self.assertEqual(
            (Address("10.1.1.3/24").get_addr()),
            addr.get_addr(),
            "Failed to update address list",
        )

        # Test 2
        with self.assertLogs("nest.topology.device", level="WARNING") as log:
            n1_n0.del_address("10.0.0.1/24")
        self.assertEqual(
            log.output,
            [
                "WARNING:nest.topology.device.device:Cannot delete an address "
                "that is not assigned. (10.0.0.1/24)"
            ],
        )

        # Test 3
        n1_n0.del_address(["10.1.1.3/24", "2001:1:1:1443::412/122"])
        addr = n1_n0.get_address(True, True, True)
        self.assertEqual(
            [],
            addr,
            "Failed to update address list",
        )

    def test_simple_lan(self):
        # pylint: disable=too-many-locals
        n0 = Node("n0")
        n1 = Node("n1")
        n2 = Node("n2")
        n3 = Node("n3")
        s0 = Switch("s0")

        (n0_s0, _) = connect(n0, s0)
        (n1_s0, _) = connect(n1, s0)
        (n2_s0, _) = connect(n2, s0)
        (n3_s0, _) = connect(n3, s0)

        nodes = [n0, n1, n2, n3]
        interface = [n0_s0, n1_s0, n2_s0, n3_s0]

        n0_s0.set_address("10.0.0.1/24")
        n1_s0.set_address("10.0.0.2/24")
        n2_s0.set_address("10.0.0.3/24")
        n3_s0.set_address("10.0.0.4/24")

        for x in nodes:
            for y in interface:
                status = x.ping(y.address, packets=1)
                if not status:
                    break

        self.assertTrue(status)

    def test_dumbbell_lan(self):
        # pylint: disable=too-many-locals
        n0 = Node("n0")
        n1 = Node("n1")
        n2 = Node("n2")
        n3 = Node("n3")
        s0 = Switch("s0")
        s1 = Switch("s1")

        (n0_s0, _) = connect(n0, s0)
        (n1_s0, _) = connect(n1, s0)
        (n2_s1, _) = connect(n2, s1)
        (n3_s1, _) = connect(n3, s1)

        connect(s0, s1)

        nodes = [n0, n1, n2, n3]
        interface = [n0_s0, n1_s0, n2_s1, n3_s1]

        n0_s0.set_address("10.0.0.1/24")
        n1_s0.set_address("10.0.0.2/24")
        n2_s1.set_address("10.0.0.3/24")
        n3_s1.set_address("10.0.0.4/24")

        for x in nodes:
            for y in interface:
                status = x.ping(y.address, packets=1)
                if not status:
                    break

        self.assertTrue(status)

    def test_tcp_param(self):
        self.n0.configure_tcp_param("ecn", "1")
        ecn = self.n0.read_tcp_param("ecn")
        self.assertEqual(ecn, "1")

    def test_invalid_interface_name(self):
        # Disable topology map
        config.set_value("assign_random_names", False)

        # Valid interface name
        Interface("namewith15chars")

        # Invalid interface name
        with self.assertRaises(ValueError) as cm:
            Interface("looonginvalidname")
        err = cm.exception
        self.assertEqual(
            str(err),
            "Device name looonginvalidname is too long. Device names "
            "should not exceed 15 characters",
        )

        # Enable topology map
        config.set_value("assign_random_names", True)

    def test_invalid_veth_name(self):
        # Disable topology map
        config.set_value("assign_random_names", False)

        node0 = Node("longname0")
        node1 = Node("longname1")

        with self.assertRaises(ValueError) as cm:
            connect(node0, node1)
        err = cm.exception
        self.assertEqual(
            str(err),
            "Auto-generated device name longname0-longname1-0 is "
            "too long. The length of name should not exceed 15 characters.",
        )

        # Enable topology map
        config.set_value("assign_random_names", True)

    def test_valid_veth_name(self):
        node0 = Node("node0")
        node1 = Node("node1")

        eth0, eth1 = connect(node0, node1)

        self.assertEqual(
            eth0.name, "node0-node1-0", "eth0 has unexpected autogenerated name"
        )
        self.assertEqual(
            eth1.name, "node1-node0-0", "eth1 has unexpected autogenerated name"
        )

        eth2, eth3 = connect(node0, node1)

        self.assertEqual(
            eth2.name, "node0-node1-1", "eth2 has unexpected autogenerated name"
        )
        self.assertEqual(
            eth3.name, "node1-node0-1", "eth3 has unexpected autogenerated name"
        )

        eth4, eth5 = connect(node0, node1, "eth4", "eth5")

        self.assertEqual(eth4.name, "eth4", "eth4 has unexpected user-given name")
        self.assertEqual(eth5.name, "eth5", "eth5 has unexpected user-given name")

    def test_invalid_ifb_interface_name(self):
        # Disable topology map
        config.set_value("assign_random_names", False)

        node0 = Node("node0")
        node1 = Node("node1")
        eth0, _ = connect(node0, node1)

        with self.assertRaises(ValueError) as cm:
            eth0.set_attributes("10mbit", "10ms", "codel")
        err = cm.exception
        self.assertEqual(
            str(err),
            "Device name ifb-node0-node1-0 is too "
            "long. Device names should not exceed 15 characters",
        )

        # Enable topology map
        config.set_value("assign_random_names", True)

    def test_run_inside_node(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        n0_n1.set_address("10.0.0.1/24")
        n1_n0.set_address("10.0.0.2/24")

        # Run ping from self.n0 to self.n1
        with self.n0:
            command = f"ping -c 1 {n1_n0.address.get_addr(with_subnet=False)}"
            with subprocess.Popen(
                command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
            ) as proc:
                (stdout, _) = proc.communicate()

        self.assertEqual(stdout[:4], b"PING", "Invalid ping output")

    def test_prp_router_api(self):
        # pylint: disable=invalid-name
        r = Router("r")

        (n0_r, r_n0) = connect(self.n0, r)
        (r_n1, n1_r) = connect(r, self.n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        self.n0.add_route("DEFAULT", n0_r)
        self.n1.add_route("DEFAULT", n1_r)

        status = self.n0.ping("10.1.2.1")

        self.assertTrue(status)

    def test_duplicate_ip_forwarding(self):
        """to enable ip forwarding more than once"""

        h1 = Node("h1")
        h2 = Node("h2")
        r1 = Node("r1")
        r1.enable_ip_forwarding(ipv6=False)
        r1.enable_ip_forwarding(ipv4=False)

        (eth1, etr1a) = connect(h1, r1)
        (etr1b, eth2) = connect(r1, h2)

        eth1.set_address("192.168.1.1/24")
        etr1a.set_address("192.168.1.2/24")

        etr1b.set_address("192.168.2.1/24")
        eth2.set_address("192.168.2.2/24")

        eth1.set_attributes("5mbit", "5ms")
        etr1b.set_attributes("5mbit", "5ms")

        eth2.set_attributes("10mbit", "100ms")
        etr1a.set_attributes("10mbit", "100ms")

        h1.add_route("DEFAULT", eth1)
        h2.add_route("DEFAULT", eth2)

        status = h1.ping(eth2.address)

        self.assertTrue(status)

        eth1.set_address("2001::1:1/122")
        etr1a.set_address("2001::1:2/122")

        etr1b.set_address("2001::2:1/122")
        eth2.set_address("2001::2:2/122")

        eth1.set_attributes("5mbit", "5ms")
        etr1b.set_attributes("5mbit", "5ms")

        eth2.set_attributes("10mbit", "100ms")
        etr1a.set_attributes("10mbit", "100ms")

        h1.add_route("DEFAULT", eth1)
        h2.add_route("DEFAULT", eth2)

        status = h1.ping(eth2.address)

        self.assertTrue(status)

    def test_enable_disable_interfaces(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        n0_n1.set_address("10.0.0.1/24")
        n1_n0.set_address("10.0.0.2/24")

        n0_n1.enable(2.0, 3.0)
        n0_n1.enable(5.0)

        with self.assertRaises(ValueError) as err:
            n1_n0.enable(-2.0, 3.0)
        self.assertEqual(
            str(err.exception),
            "start_time should be greater than or equal to 0 "
            "and end_time should be greater than 0",
        )

        with self.assertRaises(ValueError) as err:
            n1_n0.enable(4.0, 3.0)
        self.assertEqual(str(err.exception), "4.0 should be smaller than 3.0")

        with self.assertRaises(ValueError) as err:
            n1_n0.enable(end_time=4.0)
        self.assertEqual(
            str(err.exception),
            "start_time is required and should be greater than or equal to 0",
        )

        status_1 = self.n0.ping("10.0.0.2", packets=7)

        self.assertTrue(status_1)

        n0_n1.enable()
        status_2 = self.n0.ping("10.0.0.2")
        self.assertTrue(status_2)

        n0_n1.disable(1.5, 3.5)
        n0_n1.disable(5.0)

        with self.assertRaises(ValueError) as err:
            n1_n0.disable(-2.0, 4.0)
        self.assertEqual(
            str(err.exception),
            "start_time should be greater than or equal to 0 "
            "and end_time should be greater than 0",
        )

        with self.assertRaises(ValueError) as err:
            n1_n0.disable(4.0, 3.0)
        self.assertEqual(str(err.exception), "4.0 should be smaller than 3.0")

        with self.assertRaises(ValueError) as err:
            n1_n0.disable(end_time=4.0)
        self.assertEqual(
            str(err.exception),
            "start_time is required and should be greater than or equal to 0",
        )

        status_3 = self.n0.ping("10.0.0.2", packets=7)

        self.assertTrue(status_3)

        n0_n1.disable()
        status_4 = self.n0.ping("10.0.0.2")
        self.assertFalse(status_4)

    def test_two_lans_via_router(self):
        # pylint: disable=too-many-locals
        s1 = Switch("s1")
        s2 = Switch("s2")
        r1 = Router("r1")

        (n0_s1, _) = connect(self.n0, s1)
        (n1_s1, _) = connect(self.n1, s1)
        (n2_s1, _) = connect(self.n2, s1)
        (n3_s2, _) = connect(self.n3, s2)
        (n4_s2, _) = connect(self.n4, s2)
        (n5_s2, _) = connect(self.n5, s2)
        (_, r1_s1) = connect(s1, r1)
        (_, r1_s2) = connect(s2, r1)

        n0_s1.set_address("192.168.1.1/24")
        n1_s1.set_address("192.168.1.2/24")
        n2_s1.set_address("192.168.1.3/24")
        r1_s1.set_address("192.168.1.4/24")
        n3_s2.set_address("192.168.2.1/24")
        n4_s2.set_address("192.168.2.2/24")
        n5_s2.set_address("192.168.2.3/24")
        r1_s2.set_address("192.168.2.4/24")

        s1.set_address("192.168.1.5/24")
        s2.set_address("192.168.2.5/24")

        self.n0.add_route("DEFAULT", n0_s1, r1_s1.address)
        self.n1.add_route("DEFAULT", n1_s1, r1_s1.address)
        self.n2.add_route("DEFAULT", n2_s1, r1_s1.address)
        self.n3.add_route("DEFAULT", n3_s2, r1_s2.address)
        self.n4.add_route("DEFAULT", n4_s2, r1_s2.address)
        self.n5.add_route("DEFAULT", n5_s2, r1_s2.address)

        # `Ping` from `n0` to 'n3', `n1` to `n4`, and `n2` to `n5`
        status_1 = self.n0.ping(n3_s2.address)
        status_2 = self.n1.ping(n4_s2.address)
        status_3 = self.n2.ping(n5_s2.address)

        self.assertTrue(status_1)
        self.assertTrue(status_2)
        self.assertTrue(status_3)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""t_shark API management"""

import os
import logging
from .exec import exec_subprocess

logger = logging.getLogger(__name__)

# Get the current working directory
current_dir = os.getcwd()


def capture_packets(ns_name, **kwargs):
    """
    Executes packet capture using tshark

    Parameters
    ----------
    ns_name : str
        The namespace where packets need to be captured

    """

    tshark_params = ""
    packet_capture = ""
    for param, value in kwargs.items():
        if param == "-w":
            packet_capture = value

        tshark_params += f"{param} {value} "

    # Set the path for the pcap file in the current directory
    pcap_file = f"{current_dir}/packet_capture/{packet_capture}"

    # Create the directory for the pcap file if it doesn't exist
    os.makedirs(os.path.dirname(pcap_file), exist_ok=True)

    logger.info("The PCAP file is saved in %s", pcap_file)

    exec_subprocess(f"ip netns exec {ns_name} tshark {tshark_params} -w {pcap_file} -q")
    exec_subprocess(f"ip netns exec {ns_name} chmod +r {pcap_file}")



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""Test APIs from topology sub packages"""

import unittest
from nest.topology import Node, connect
from nest.clean_up import delete_namespaces
from nest.topology_map import TopologyMap

# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=too-many-arguments


class TestTopologyPingPreload(unittest.TestCase):
    def setUp(self):
        self.n0 = Node("n0")
        self.n1 = Node("n1")

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()

    def test_p2p_ping_preload(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        n0_n1.set_address("10.0.0.1/24")
        n1_n0.set_address("10.0.0.2/24")

        status = self.n0.ping("10.0.0.2", preload=10, packets=10)

        self.assertTrue(status)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""MPEG-DASH Encoder"""

from pathlib import Path
import logging
import re
import shutil

from nest.engine.exec import exec_exp_commands, exec_subprocess_with_live_output
from nest.engine.util import is_dependency_installed

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# pylint: disable=too-few-public-methods
class MpegDashEncoder:
    """
    Contains the configuration settings and script to encode a video.
    """

    mpeg_dash_encoded_chunks_path_list = []

    def __init__(
        self,
        gop_size: int = 15,
        fps: int = 30,
        video_codec: str = "libx264",
        pixel_format: str = "yuv420p",
        audio_codec: str = "aac",
        audio_bitrate: str = "128k",
        segment_duration: int = 1,
        aspect_ratio: str = "16/9",
        video_stream_info: list = None,
    ):
        """
        Constructor to initialize the MPEG-DASH encoder

        Parameters
        ----------
        gop_size : int
            Group of pictures (GOP) size
        fps : int
            Frame rate (per second)
        video_codec: str
            Name of the video codec to be used
        pixel_format: str
            Type of pixel format (Supported formats can be viewed
            here: https://github.com/gpac/gpac/wiki/filters_properties#pixel-formats)
        audio_codec: str
            Name of the audio codec to be used
        audio_bitrate: str
            Amount of audio bitrate required
        segment_duration: int
            Duration of each encoded MPEG-DASH segment
        aspect_ratio: str
            Ratio between the video's width and height.
        video_stream_info: list
            List containing JSON objects that contain video resolution and bitrate.
        """
        if not is_dependency_installed("gpac"):
            raise RuntimeError("Encoding Failed! gpac is not installed!")

        if not isinstance(gop_size, int):
            raise ValueError("gop_size is supposed to be an integer value.")

        if not isinstance(fps, int):
            raise ValueError("fps is supposed to be an integer value.")

        if not isinstance(segment_duration, int):
            raise ValueError("segment_duration is supposed to be an integer value")

        if video_stream_info is None:
            video_stream_info = [
                {"resolution": "256x144", "bitrate": "50K"},
                {"resolution": "427x240", "bitrate": "100K"},
                {"resolution": "640x360", "bitrate": "400K"},
                {"resolution": "854x480", "bitrate": "750K"},
                {"resolution": "1280x720", "bitrate": "2.5M"},
                {"resolution": "1920x1080", "bitrate": "6M"},
            ]

        resolution_pattern = r"\d+x\d+"
        bitrate_pattern = r"^\d+(\.\d+)?[KkMm]"
        aspect_ratio_pattern = r"^\d+/\d+$"

        if not re.match(bitrate_pattern, audio_bitrate):
            raise ValueError(
                '"bitrate" for audio is not in the specified format. i.e. "2.5M" or "5K".'
            )

        if not re.match(aspect_ratio_pattern, aspect_ratio):
            raise ValueError(
                'Aspect Ratio is not in the specified format. i.e. "<width>/<height>". E.g. "16/9"'
            )

        for video_repr in video_stream_info:
            if ("resolution" not in video_repr) or ("bitrate" not in video_repr):
                raise KeyError('"resolution" or "bitrate" key not found')

            if not re.match(resolution_pattern, video_repr["resolution"]):
                raise ValueError(
                    '''"resolution" is not in the specified format.
                    i.e. "<width>x<height>". E.g. "1920x1080"'''
                )

            if not re.match(bitrate_pattern, video_repr["bitrate"]):
                raise ValueError(
                    '"bitrate" for video is not in the specified format. i.e. "2.5M" or "5K".'
                )

        self.gop_size = gop_size
        self.fps = fps
        self.video_codec = video_codec
        self.pixel_format = pixel_format
        self.audio_codec = audio_codec
        self.audio_bitrate = audio_bitrate
        self.segment_duration = segment_duration
        self.aspect_ratio = aspect_ratio
        self.video_stream_info = video_stream_info

    def encode_video(
        self, input_video_path: Path, output_path: Path, overwrite: bool = False
    ):
        """
        Encodes the file corresponding to `input_video_path` and
        stores the encoded chunks in the `output_path`.

        Parameters
        ----------
        input_video_path: Path
            Path of the input video file
        output_path: Path
            Path to store the encoded chunks in.
        overwrite: bool
            If 'True' then the 'output_path' directory is emptied and then chunks are stored.

        Returns
        -------
        int
            Return code recieved after the encoding process terminates
        """

        if not Path.exists(input_video_path):
            logger.warning("The path '%s' does not exist.", str(input_video_path))
            logger.warning("Using a sample video from the internet for encoding.")

            if not is_dependency_installed("wget"):
                raise RuntimeError(
                    "Error downloading sample video! wget is not installed!"
                )

            input_video_url = (
                "https://commondatastorage.googleapis.com/"
                "gtv-videos-bucket/sample/ForBiggerEscapes.mp4"
            )

            download_dir = "/tmp/mpeg-dash-sample-video/"
            exec_subprocess_with_live_output(
                f"wget {input_video_url} -P {download_dir} -nc -q --show-progress"
            )

            input_video_path = (
                download_dir + input_video_url.rsplit("/", maxsplit=1)[-1]
            )

        MpegDashEncoder.mpeg_dash_encoded_chunks_path_list.append(output_path)

        if Path.exists(output_path) and any(
            file.suffix == ".mpd" for file in output_path.iterdir()
        ):
            if overwrite is False:
                logger.warning(
                    "The path '%s' already contains an MPD file.", str(output_path)
                )
                logger.warning(
                    "Utilizing the pre-existing MPD file for MPEG-DASH streaming."
                )
                return 0
            shutil.rmtree(output_path)
            logger.warning(
                "Overwriting the files in the directory '%s' .", str(output_path)
            )
        else:
            output_path.mkdir(parents=True, exist_ok=True)

        cmd_string = f" gpac -i {input_video_path}:FID=srcFile:FPS={self.fps} "

        # Setting video encoding parameters
        for i in range(len(self.video_stream_info)):
            cmd_string += f"ffsws:osize={self.video_stream_info[i]['resolution']}"
            cmd_string += f":osar={self.aspect_ratio}"
            cmd_string += (
                f":SID=srcFile:FID={self.video_stream_info[i]['resolution']}_res"
            )
            cmd_string += f" enc:c={self.video_codec}"
            cmd_string += f":fintra=1:bitrate={self.video_stream_info[i]['bitrate']}"
            cmd_string += f":cgop=true:gop={self.gop_size}"
            cmd_string += (
                f":SID={self.video_stream_info[i]['resolution']}_res:FID=EV{i} "
            )

        # Setting audio encoding parameters
        cmd_string += f"c={self.audio_codec}:SID=srcFile"
        cmd_string += f":FID=EA:bitrate={self.audio_bitrate} "
        cmd_string += f"-o {output_path}/manifest.mpd:gpac"
        cmd_string += ":SID="
        for i in range(len(self.video_stream_info)):
            cmd_string += f"EV{i},"
        cmd_string += f"EA:segdur={self.segment_duration}:cdur=0.5"
        cmd_string += (
            ":profile=dashavc264.live:title='MPEG-DASH Application (NITK-NeST)'"
        )

        logger.info(
            "Video encoding started. Please be patient. This may take a while.."
        )

        status = exec_exp_commands(cmd_string)

        if status == 0:
            logger.info("Video encoding completed successfully!")
        else:
            logger.error("Video encoding failed!")

        return status



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2023 NITK Surathkal

"""Functions related to ARP table"""

import re
import logging
from .exec import exec_subprocess_with_live_output

logger = logging.getLogger(__name__)


def get_arp_table(ns_name):
    """
    Retrieve and print the ARP table for a specified namespace.

    Parameters
    ----------
    ns_name : str
        The name of the namespace for which to retrieve the ARP table.

    Returns
    -------
    bool
        Returns `True` if the ARP table was successfully retrieved and printed,
        Else `False`
    """
    status = exec_subprocess_with_live_output(f"ip netns exec {ns_name} arp")
    return status == 0


def delete_arp_entry(ns_name, ip_addr):
    """
    Delete a specified ARP table entry for a namespace.

    Parameters
    ----------
    ns_name : str
        The name of the namespace in which to delete the ARP table entry.
    ip_addr : Address
        The IP address of the ARP table entry to be deleted.

    Returns
    -------
    bool
        Returns `True` if the ARP table entry was successfully deleted,
        Else `False`
    """

    status = exec_subprocess_with_live_output(
        f"ip netns exec {ns_name} arp -d {ip_addr}"
    )
    return status


def flush_arp_table(ns_name):
    """
    Flush the entire ARP table for a specified namespace.

    Parameters
    ----------
    ns_name : str
        The name of the namespace for which to flush the ARP table.

    Returns
    -------
    bool
        Returns True if the ARP table was successfully flushed,
        Else `False`
    """
    status = exec_subprocess_with_live_output(
        f"ip netns exec {ns_name} ip -s neigh flush all"
    )
    return status


def get_arp_entry(ns_name, ip_addr):
    """
    Retrieve a specified ARP table entry for a namespace.

    Parameters
    ----------
    ns_name : str
        The name of the namespace from which to retrieve the ARP table entry.
    ip_addr : Address
        The IP address of the ARP table entry to be retrieved.

    Returns
    -------
    bool
        Returns `True` if the ARP table entry was successfully retrieved
        Else `False`
    """
    status = exec_subprocess_with_live_output(f"ip netns exec {ns_name} arp {ip_addr}")
    return status


def set_arp_entry(ns_name, ip_addr, mac_addr):
    """
    Set the ARP table entry for a specified IP address and MAC address from the given node and
    Validate MAC addresses.

    Parameters
    ----------
    ns_name : str
        The name of the namespace from which to retrieve the ARP table entry.
    ip_addr : Address
        The IP address of the ARP table entry to be retrieved.
    mac_address : string
            The MAC address for which to set the ARP table entry.

    Returns
    -------
    bool
        Returns `True` if the ARP table entry was successfully retrieved
        Else `False`
    """
    # regular expression pattern
    mac_address_pattern = "^([0-9A-Fa-f]{2}[:]){5}([0-9A-Fa-f]{2})$"
    # regular expression pattern matches a MAC address
    if re.match(mac_address_pattern, mac_addr):
        status = exec_subprocess_with_live_output(
            f"ip netns exec {ns_name} arp -s {ip_addr} {mac_addr}"
        )
    else:
        logger.info("%s is not a valid Mac address", mac_addr)
    return status



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

""" Constructs CoAP client and sends GET/PUT requests to the server """

# Below implementation is based on the examples from:
# https://aiocoap.readthedocs.io/en/latest/examples.html

import sys
import argparse
import asyncio
import json
import time

import aiocoap


async def get_req(context, uri, mtype, non_timeout):
    """
    Send a GET request to the CoAP server to the given URI

    Parameters
    ----------
    context : aiocoap.Context
        The context object for the client application
    uri : str
        The URI of the resource to which the request is to be made
    mtype : str
        Confirmable(CON) or non-confirmable(NON) message
    non_timeout : int
        Timeout to wait for responses to NON messages
    """
    # Needed to store the output and error data
    out_dict = {}
    err_dict = {}

    # Choosing between confirmable(CON) and non-confirmable(NON) message
    if mtype == "CON":
        # pylint: disable=no-member
        request = aiocoap.Message(code=aiocoap.GET, mtype=aiocoap.Type.CON, uri=uri)
    elif mtype == "NON":
        # pylint: disable=no-member
        request = aiocoap.Message(code=aiocoap.GET, mtype=aiocoap.Type.NON, uri=uri)

    try:
        # Start time for measuring RTT
        rtt_start = time.time()

        # For NON messages, we need a timeout for the response since it is not possible
        # to keep waiting for it. For CON messages, the timeout is set and managed
        # internally in aiocoap so we need not handle it.
        if mtype == "CON":
            # Send request and wait for response
            response = await context.request(request).response
        else:
            # Send request and wait for response or timeout
            response = await asyncio.wait_for(
                context.request(request).response, timeout=non_timeout
            )

        # RTT calculation
        rtt = time.time() - rtt_start

    # Detect asyncio timeout errors for NON messages
    except asyncio.TimeoutError:
        err_dict = {"Error": "Timeout Error", "Exception": str(asyncio.TimeoutError)}
    # Any other misc errors
    # pylint: disable=broad-except
    except Exception as exception:
        err_dict = {"Error": "Failed to fetch resource", "Exception": str(exception)}
    else:
        # Convert raw bytes to the "c.dd" form
        response_code = response.code.dotted

        # Convert the payload in bytes to string
        response_payload_bytes = response.payload
        response_payload = response_payload_bytes.decode("utf-8")

        out_dict = {
            "Response Code": response_code,
            "Response Payload": response_payload,
            "Message Type": mtype,
            "RTT(in secs)": rtt,
        }

    return (out_dict, err_dict)


# pylint: disable=too-many-locals
async def put_req(context, uri, message_payload, mtype, non_timeout):
    """
    Perform a PUT request to the given URI.

    Parameters
    ----------
    context : aiocoap.Context
        The context object for the client application
    uri : str
        The URI of the resource to which the request is to be made
    message_payload : str
        Message payload sent to the server via the PUT request
    mtype : str
        Confirmable(CON) or non-confirmable(NON) message
    non_timeout : int
        Timeout to wait for responses to NON messages
    """
    # Needed to store the output and error data
    out_dict = {}
    err_dict = {}

    payload = message_payload.encode("utf-8")

    # Choosing between confirmable(CON) and non-confirmable(NON) message
    if mtype == "CON":
        # pylint: disable=no-member
        request = aiocoap.Message(
            code=aiocoap.PUT, mtype=aiocoap.Type.CON, payload=payload, uri=uri
        )
    elif mtype == "NON":
        # pylint: disable=no-member
        request = aiocoap.Message(
            code=aiocoap.PUT, mtype=aiocoap.Type.NON, payload=payload, uri=uri
        )

    try:
        # Start time for measuring RTT
        rtt_start = time.time()

        # For NON messages, we need a timeout for the response since it is not possible
        # to keep waiting for it. For CON messages, the timeout is set and managed
        # internally in aiocoap so we need not handle it.
        if mtype == "CON":
            # Send request and wait for response
            response = await context.request(request).response
        else:
            # Send request and wait for response or timeout
            response = await asyncio.wait_for(
                context.request(request).response, timeout=non_timeout
            )

        # RTT calculation
        rtt = time.time() - rtt_start

    # Detect asyncio timeout errors for NON messages
    except asyncio.TimeoutError:
        err_dict = {"Error": "Timeout Error", "Exception": str(asyncio.TimeoutError)}
    # Any other misc errors
    # pylint: disable=broad-except
    except Exception as exception:
        err_dict = {"Error": "Failed to fetch resource", "Exception": str(exception)}
    else:
        # Convert raw bytes to the "c.dd" form
        response_code = response.code.dotted

        # Convert the payload in bytes to string
        response_payload_bytes = response.payload
        response_payload = response_payload_bytes.decode("utf-8")

        out_dict = {
            "Response Code": response_code,
            "Response Payload": response_payload,
            "Message Type": mtype,
            "RTT(in secs)": rtt,
        }

    return (out_dict, err_dict)


# pylint: disable=too-many-locals
async def main():
    """
    Parses client request parameters and generates appropriate traffic
    """

    help_description = "CoAP Client Application"

    # Initialize parser
    parser = argparse.ArgumentParser(description=help_description)

    # Adding arguments for client request
    parser.add_argument("-r", "--request_type", type=str, choices=["put", "get"])
    parser.add_argument("-m", "--message_payload", type=str)
    parser.add_argument("-d", "--destination_ip", type=str)
    parser.add_argument("-c", "--n_con_msgs", type=str)
    parser.add_argument("-n", "--n_non_msgs", type=str)
    parser.add_argument("-t", "--non_timeout", type=str)

    # Read Arguments
    args = parser.parse_args()

    # Constructing the URI from the destination IP
    uri = "coap://" + args.destination_ip + "/sample-resource"

    # Create client context
    context = await aiocoap.Context.create_client_context()

    # Lists for storing outputs and errors for every packet sent
    outputs = []
    errors = []

    total_msgs = int(args.n_non_msgs) + int(args.n_con_msgs)

    # List of coroutine to pass to `gather` for concurrent execution
    coroutine_list = []

    # Parse arguments
    for i in range(0, total_msgs):
        if i < int(args.n_non_msgs):
            mtype = "NON"
        else:
            mtype = "CON"

        # Add GET and PUT request coroutines to the coroutine list
        if args.request_type == "get":
            coroutine_list.append(get_req(context, uri, mtype, float(args.non_timeout)))
        else:
            coroutine_list.append(
                put_req(
                    context, uri, args.message_payload, mtype, float(args.non_timeout)
                )
            )

    # Start time for measuring Flow Completion Time(FCT)
    fct_start = time.time()

    # Run all the coroutines
    results = await asyncio.gather(*coroutine_list)

    # Calculate FCT
    # This includes NON messages as well since they can receive responses as well
    fct = time.time() - fct_start

    # Add the FCT to the outputs dictionary
    outputs.append({"Flow Completion Time(in secs)": fct})

    # Split the final results into outputs and errors
    for i in results:
        (output, error) = i
        outputs.append(output)
        errors.append(error)

    # Convert the output and error lists into JSON format
    outputs_json = json.dumps(outputs)
    errors_json = json.dumps(errors)

    # Print outputs and errors into respective streams
    print(outputs_json, file=sys.stdout)
    print(errors_json, file=sys.stderr)


if __name__ == "__main__":
    asyncio.get_event_loop().run_until_complete(main())



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""Test APIs for iperf3 server options"""

import json
import os
import unittest
from nest.experiment.pack import Pack

from nest.topology import Node, connect, Switch
from nest.clean_up import delete_namespaces
from nest.topology_map import TopologyMap
from nest.experiment import Experiment, Flow
from nest.experiment.tools import Iperf3


# pylint: disable=unused-variable
def search_file(filename):
    """
    Search file in current directory

    Parameter:
    ----------
    filename: str
        file name to search

    Returns:
        if found return file else False
    """
    folder_name = Pack.FOLDER
    folder_path = os.path.join(os.getcwd(), folder_name)
    file_name = filename
    if os.path.exists(folder_path):
        for _, _, files in os.walk(folder_path):
            if file_name in files:
                filepath = os.path.join(folder_path, file_name)
                with open(filepath, "r") as file:
                    result = json.loads(file.read())
                    file.close()
                return result
    return False


# pylint: disable=invalid-name, unused-variable, redefined-builtin
# pylint: disable=too-many-instance-attributes,  too-many-nested-blocks
class TestIperf3Options(unittest.TestCase):
    """
    Test cases for different options enabled in iperf3
    """

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()

    def setUp(self):
        # Network setup to test different iperf3 options
        # Local Area Networks (LANs) connected directly to
        # each other. LAN-1 consists three hosts `h1` to `h3` connected to switch `s1`,
        # and LAN-2 consists three hosts `h4` to `h6` connected to switch `s2`.
        # Switches `s1` and `s2` are connected to each other.

        #########################################################
        #                    Network Topology                   #
        #           LAN-1                      LAN-2            #
        #   h1 ---------------            ---------------- h4   #
        #                     \         /                       #
        #   h2 --------------- s1 ---- s2 ---------------- h5   #
        #                     /         \                       #
        #   h3 ---------------            ---------------- h6   #
        #             <------ 100mbit, 1ms ------>              #
        #                                                       #
        #########################################################

        # Create six hosts `h1` to `h6`, and two switches `s1` and `s2`
        self.h1 = Node("h1")
        self.h2 = Node("h2")
        self.h3 = Node("h3")
        self.h4 = Node("h4")
        self.h5 = Node("h5")
        self.h6 = Node("h6")
        self.s1 = Switch("s1")
        self.s2 = Switch("s2")

        # Create LAN-1: Connect hosts `h1`, `h2` and `h3` to switch `s1`
        # `eth1` to `eth3` are the interfaces at `h1` to `h3`, respectively.
        # `ets1a` is the first interface at `s1` which connects it with `h1`
        # `ets1b` is the second interface at `s1` which connects it with `h2`
        # `ets1c` is the third interface at `s1` which connects it with `h3`
        (eth1, ets1a) = connect(self.h1, self.s1)
        (eth2, ets1b) = connect(self.h2, self.s1)
        (eth3, ets1c) = connect(self.h3, self.s1)

        # Create LAN-2: Connect hosts `h4`, `h5` and `h6` to switch `s2`
        # `eth4` to `eth6` are the interfaces at `h4` to `h6`, respectively.
        # `ets2a` is the first interface at `s2` which connects it with `h4`
        # `ets2b` is the second interface at `s2` which connects it with `h5`
        # `ets2c` is the third interface at `s2` which connects it with `h6`
        (eth4, ets2a) = connect(self.h4, self.s2)
        (eth5, ets2b) = connect(self.h5, self.s2)
        (eth6, ets2c) = connect(self.h6, self.s2)

        # Connect switches `s1` and `s2`
        # `ets1d` is the fourth interface at `s1` which connects it with `s2`
        # `ets2d` is the fourth interface at `s2` which connects it with `s1`
        (ets1d, ets2d) = connect(self.s1, self.s2)

        # Assign IPv4 addresses to all the interfaces.
        # We assume that the IPv4 address of this network is `192.168.1.0/24`.
        # Note: IP addresses should not be assigned to the interfaces on the switches.
        eth1.set_address("192.168.1.1/24")
        eth2.set_address("192.168.1.2/24")
        eth3.set_address("192.168.1.3/24")
        eth4.set_address("192.168.1.4/24")
        eth5.set_address("192.168.1.5/24")
        eth6.set_address("192.168.1.6/24")

        # Set the link attributes
        eth1.set_attributes("100mbit", "1ms")
        eth2.set_attributes("100mbit", "1ms")
        eth3.set_attributes("100mbit", "1ms")
        eth4.set_attributes("100mbit", "1ms")
        eth5.set_attributes("100mbit", "1ms")
        eth6.set_attributes("100mbit", "1ms")

        # Assign  source, Destination, start time, end time and number of
        # parallel flows to each udp flows
        self.flow_udp_1 = Flow(self.h1, self.h4, eth4.get_address(), 0, 5, 1)
        self.flow_udp_2 = Flow(self.h2, self.h4, eth4.get_address(), 5, 15, 1)

    def test_set_dport(self):
        """
        In this test case user defines a port no to start server on the destination node.
        At the end of experiment, we will searh for destination port no in the output file.
        if destination port no and user defined port no are same, that means the test case
        has successfully passed.
        """
        portno = 1234

        # iperf3_server_options API is used to configure iperf3 server options
        exp = Experiment("flow-udp")
        exp.add_udp_flow(
            self.flow_udp_1,
            server_options=Iperf3.server_option(s_interval=0.5, port_no=portno),
            client_options=Iperf3.client_option(interval=0.4, cport=12345),
        )
        exp.run()

        exp_output = search_file("iperf3Server.json")
        dst_name = self.h4.name
        src_ip = "192.168.1.1"
        status = False
        if exp_output:
            for dict in exp_output[dst_name]:
                if src_ip in dict:
                    for dst_port in dict[src_ip]:
                        if portno == int(dst_port):
                            status = True
                        else:
                            print(
                                f"Failed! input port_no {portno} is not equal "
                                "to server port_no {dst_port}"
                            )
        self.assertTrue(status)

    def test_set_interval(self):
        """
        In this test case user defines a time interval for each output line at the server.
        At the end of experiment, we will searh for value of "duration" in the output JSON file.
        if the average of "duration" and user defined interval is same, that means this
        test case has successfully passed.
        """

        # iperf3_server_options API is used to configure iperf3 server options
        interval = 0.4
        portno = 1234

        exp = Experiment("flow")
        tool_options = {"s_interval": interval, "port_no": portno}
        exp.add_udp_flow(self.flow_udp_1, server_options=tool_options)
        exp.run()

        exp_output = search_file("iperf3Server.json")
        dst_name = self.h4.name
        src_ip = "192.168.1.1"
        status = False
        if exp_output:
            for dict in exp_output[dst_name]:
                if src_ip in dict:
                    interval_val = []
                    for dst_port in dict[src_ip]:
                        for dict1 in dict[src_ip][dst_port]:
                            if "duration" in dict1:
                                interval_val.append(float(dict1["duration"]))
                    average = round(sum(interval_val) / len(interval_val), 1)
                    if interval == average:
                        status = True
                    else:
                        print(
                            f"Failed! input interval {interval} is not equal "
                            "to output interval {average}"
                        )

        self.assertTrue(status)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2023 NITK Surathkal

"""Congestion Algorithm list for TCP/MPTCP Experiments"""

congestion_algo_list = [
    "bbr",
    "bic",
    "cdg",
    "cubic",
    "dctcp",
    "highspeed",
    "htcp",
    "illinois",
    "reno",
    "scalable",
    "vegas",
    "veno",
    "westwood",
    "yeah",
]



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to creation of virtual ethernet devices in topology"""

import logging
from .device import Device

logger = logging.getLogger(__name__)


class VethEnd(Device):
    """
    This is the end of the veth created, to connect two devices

    Attributes
    ----------
    id : str
        This value is used by `engine` to create emulated interface
        entity
    node_id : str
        id of the Node to which this Device belongs
    """

    def __init__(self, name, node_id):
        """
        Constructor of VethEnd

        Parameters
        ----------
        name : str
            Name of the interface
        node_id : str
            Id of the node that the interface belongs to
        """

        super().__init__(name, node_id)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
API implementation of `connect`.
"""

from nest import config
from nest.input_validator import input_validator
from nest.topology import Switch
from nest.topology.interface import create_veth_pair
from nest.topology.network import Network
from nest.topology import Node
from nest.topology_map import TopologyMap

# Max length of interface when Topology Map is disabled
# i.e. 'assign_random_names' is set to False in config
MAX_CUSTOM_NAME_LEN = 15


@input_validator
def connect(
    node1: Node,
    node2: Node,
    interface1_name: str = "",
    interface2_name: str = "",
    network: Network = None,
):
    """
    Connects two nodes `node1` and `node2`.
    Creates two paired Virtual Ethernet interfaces (veth) and returns
    them as a 2-element tuple.
    The first interface belongs to `node1`, and the second interface
    belongs to `node2`.

    Parameters
    ----------
    node1 : Node
        First veth interface added in this node
    node2 : Node
        Second veth interface added in this node
    interface1_name : str
        Name of first veth interface
    interface2_name : str
        Name of second veth interface
    network : Network
        Object of the Network to add interfaces

    Returns
    -------
    (interface1, interface2)
        2 tuple of created (paired) veth interfaces. `interface1` is in
        `n1` and `interface2` is in `n2`.
    """
    # Number of connections between `node1` and `node2`, set to `None`
    # initially since it hasn't been computed yet
    connections = None

    # Check interface names
    if interface1_name == "":
        connections = _number_of_connections(node1, node2)
        interface1_name = _autogenerate_interface_name(node1, node2, connections)

    if interface2_name == "":
        if connections is None:
            connections = _number_of_connections(node1, node2)
        # pylint: disable=arguments-out-of-order
        interface2_name = _autogenerate_interface_name(node2, node1, connections)

    # Create 2 interfaces
    (interface1, interface2) = create_veth_pair(interface1_name, interface2_name)

    # pylint: disable=protected-access
    node1._add_interface(interface1)
    node2._add_interface(interface2)

    interface1.set_mode("UP")
    interface2.set_mode("UP")

    # Disabling Duplicate Address Detection(DAD) at the interfaces
    if config.get_value("disable_dad") is True:
        interface1.disable_ip_dad()
        interface2.disable_ip_dad()

    # The network parameter takes precedence over "global" network level
    if network is None:
        network = Network.current_network

    # Add the interfaces to the network if mentioned
    # For switch, add the switch interface to the network(not veth end) and decrement the
    # count of orphan interface to remove veth end from the list of orphan interfaces
    if network is not None:
        if isinstance(node1, Switch):
            network.add_interface(node1)
            TopologyMap.decrement_orphan_interfaces()
        else:
            network.add_interface(interface1)

        if isinstance(node2, Switch):
            network.add_interface(node2)
            TopologyMap.decrement_orphan_interfaces()
        else:
            network.add_interface(interface2)

    return (interface1, interface2)


def _autogenerate_interface_name(node1, node2, connections):
    """
    Auto-generate interface names based on respective node names
    and number of connections
    """
    interface_name = node1.name + "-" + node2.name + "-" + str(connections)

    if config.get_value("assign_random_names") is False:
        if len(interface_name) > MAX_CUSTOM_NAME_LEN:
            raise ValueError(
                f"Auto-generated device name {interface_name} is too long. "
                f"The length of name should not exceed 15 characters."
            )

    return interface_name


def _number_of_connections(node1, node2):
    """
    Return number of connections between the two nodes
    """
    connections = 0

    if len(node1.interfaces) > len(node2.interfaces):
        node1, node2 = node2, node1

    for interface in node1.interfaces:
        if interface.pair.node_id == node2.id:
            connections = connections + 1

    return connections



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""
Runs commands to setup SIP experiment and collect data
"""

from functools import partial
import logging

from nest.topology_map import TopologyMap
from ..results import SipResults
from .runnerbase import Runner
from ...engine.sip import run_sip_client, run_sip_server

logger = logging.getLogger(__name__)

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-few-public-methods


class SipRunner(Runner):
    """
    Runs SIP client and server.

    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        ns_id,
        destination_ip,
        dst_ns,
        port,
        duration,
        scenario,
        server_xml_file,
        client_xml_file,
        callrate,
    ):
        """
        Constructor to initialize the runner

        Parameters
        ----------
        ns_id : str
            network namespace of the client
        destination_ip : str
            the ip of server to which it has to connect
        dst_ns : str
            network namespace of the server
        port: int
            the port number of the server at which it is running

        """
        self.port = port
        self.duration = duration
        self.scenario = scenario
        self.server_xml_file = server_xml_file
        self.client_xml_file = client_xml_file
        self.callrate = callrate

        super().__init__(
            ns_id,
            start_time=None,
            run_time=None,
            destination_ip=destination_ip,
            dst_ns=dst_ns,
            use_named_out_file=True,
        )

    @staticmethod
    def run_server(ns_id, port, scenario, xml_file):
        """
        Run SIP server

        Parameters
        ----------
        ns_id : str
            namespace to run SIP server on
        port : int
            port to run SIP server on
        scenario : str
            Specifies the scenario to be run for the experiment.
            (uses "basic" by default)
        xml_file : Path
            Specifies the path to an XML scenario file for the SIPp test.
            (Used only when scenario is "xml")
        """

        return_code = run_sip_server(ns_id, port, scenario, xml_file)
        if return_code not in (99, 0):
            ns_name = TopologyMap.get_node(ns_id).name
            logger.error("Error running SIP server at %s.", ns_name)

    # pylint: disable=too-many-statements, too-many-branches
    def run(self):
        """
        Calls engine method to run SIP client
        """

        # Run SIP client
        super().run(
            partial(
                run_sip_client,
                self.ns_id,
                self.destination_address.get_addr(with_subnet=False),
                self.port,
                self.duration,
                scenario=self.scenario,
                xml_file=self.client_xml_file,
                callrate=self.callrate,
            ),
            error_string_prefix="Running SIP",
        )

    def parse(self):
        """
        Parse SIP output from self.out
        """
        # Rewind to start of the out file and read output
        self.out.seek(0)
        lines = self.out.readlines()

        data = []
        headers = lines[0].decode().split(";")

        for line in lines[1:]:
            values = line.decode().split(";")
            values.remove("\n")
            data.append(dict(zip(headers, values)))

        SipResults.add_result(
            self.ns_id,
            data[-1],
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""TCP module commands"""
import logging
from .exec import exec_subprocess

logger = logging.getLogger(__name__)

# pylint: disable=line-too-long
def is_module_loaded(cong_algo):
    """
    util to check if a module is loaded

    Parameters
    ----------
    cong_algo : str
        TCP congestion algorithm

    Returns
    -------
    bool
        true if the `TCP module` is loaded
    """
    if (
        cong_algo
        in exec_subprocess(
            "sysctl -n net.ipv4.tcp_available_congestion_control",
            output=True,
        ).split()
    ):
        return True
    return False


def set_tcp_params(cong_algo, tcp_module_params, reset_flag=False):
    """
    Change TCP module parameters

    Parameters
    ----------
    cong_algo : str
        TCP congestion algorithm
    tcp_module_params : dict
        TCP module parameters
    reset_flag : boolean
        Set or reset the tcp parameters
    """
    for key, value in tcp_module_params.items():
        if (
            exec_subprocess(
                f"ls -l /sys/module/tcp_{cong_algo}/parameters/{key}",
                output=True,
            )[2]
            == "w"
        ):
            exec_subprocess(
                f"echo -n {value} > /sys/module/tcp_{cong_algo}/parameters/{key}",
                shell=True,
            )
        elif not reset_flag:
            default_value = exec_subprocess(
                f"cat /sys/module/tcp_{cong_algo}/parameters/{key}",
                output=True,
            )
            logger.warning(
                "%s:%s parameter not modifiable. Parameter set to default value - %s",
                cong_algo,
                key,
                default_value,
            )


def load_tcp_module(cong_algo, params_string):
    """
    Load TCP module with parameters

    Parameters
    ----------
    cong_algo : str
        TCP congestion algorithm
    params_string : str
        TCP module parameters
    """
    exec_subprocess(f"modprobe tcp_{cong_algo} {params_string}")


def get_current_params(cong_algo):
    """
    util to get the current parameters of a loaded TCP module

    Parameters
    ----------
    cong_algo : str
        TCP congestion algorithm

    Returns
    -------
    dict
        current parameters
    """
    params = {}
    param_names = exec_subprocess(
        f"ls /sys/module/tcp_{cong_algo}/parameters/",
        output=True,
    ).split()
    for param_name in param_names:
        param_value = exec_subprocess(
            f"cat /sys/module/tcp_{cong_algo}/parameters/{param_name}",
            shell=True,
            output=True,
        ).strip()
        params[param_name] = param_value
    return params


def remove_tcp_module(cong_algo):
    """
    Remove congestion algorithm TCP module

    Parameters
    ----------
    cong_algo : str
        TCP congestion algorithm
    """
    exec_subprocess(f"modprobe -r tcp_{cong_algo}")



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
This folder contains classes for various network devices
in Linux, for eg., Veth, Ifb
"""

from .device import Device
from .ifb import Ifb
from .veth_end import VethEnd



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2023 NITK Surathkal

"""Validation helper for TCP Flows and Experiments"""
from colorama import Fore, Style
from nest.experiment.constants.tcp_congestion_control_algorithms import (
    congestion_algo_list,
)


class TCPValidations:
    """
    Validates vaious aspects of TCP experiments
    """

    SUCCESS_LOG = Fore.GREEN + "\N{check mark}"
    FAILURE_LOG = Fore.RED + "\N{cross mark}"
    NEUTRAL_LOG = Style.RESET_ALL + "\N{bullet}"

    def __log_success(self, msg: str):
        print(self.SUCCESS_LOG, msg)

    def __log_failure(self, msg: str):
        print(self.FAILURE_LOG, msg)

    def __log_neutral(self, msg: str):
        print(self.NEUTRAL_LOG, msg)

    def verify_tool(self, tool):
        """
        Verifies the tool to be either of iperf3 or netperf

        Parameters
        ----------
        tool : string
            Tool selected for TCP experiment
        """
        if tool not in ["iperf3", "netperf"]:
            raise ValueError(
                f"{tool} is not a valid performance tool. Should be either netperf/iperf3"
            )

    def verify_congestion_control_algorithm(self, congestion_algorithm):
        """
        Verifies the congestion control algorithm to be
        one from the congestion algorithm list

        Parameters
        ----------
        congestion_algorithm : string
            Congestion control algorithm selected for the experiment
        """
        if congestion_algorithm not in congestion_algo_list:
            raise ValueError(
                f"{congestion_algorithm} is not a valid TCP Congestion Control algorithm"
            )

    # pylint: disable=too-many-branches
    def verify_mptcp_setup(self, experiment):
        """
        This function prepares a checklist that tells the user whether their
        topology is MPTCP enabled.

        Parameters
        ----------
        experiment: Experiment
            Experiment object that may contain some flows for the experiment
        """

        print(Style.BRIGHT, "\nMPTCP Logger Invoked", Style.RESET_ALL)

        # Check if there is atleast one MPTCP flow in the experiment
        if not any(map(lambda flow: flow.protocol == "MPTCP", experiment.flows)):
            self.__log_failure("No MPTCP Flows found in this experiment")
        else:
            self.__log_success("MPTCP Flows exist in this experiment")

        print(Style.RESET_ALL)

        # Loop over every flow
        for flow in filter(lambda x: x.protocol == "MPTCP", experiment.flows):
            self.__log_neutral(f"{Style.BRIGHT}{flow}{Style.RESET_ALL}")

            # Source Node must be MPTCP enabled
            if flow.source_node.is_mptcp:
                self.__log_success(
                    f"Source Node {flow.source_node.name} is MPTCP enabled"
                )
            else:
                self.__log_failure(
                    f"Source Node {flow.source_node.name} is not MPTCP enabled"
                )

            # Destination Node must be MPTCP enabled
            if flow.destination_node.is_mptcp:
                self.__log_success(
                    f"Destination Node {flow.destination_node.name} is MPTCP enabled"
                )
            else:
                self.__log_failure(
                    f"Destination Node {flow.destination_node.name} is not MPTCP enabled"
                )

            # Either the source or the destination must be multi-homed and multi-addressed
            if (
                flow.source_node.is_mptcp_supported()
                or flow.destination_node.is_mptcp_supported()
            ):
                self.__log_success(
                    f"Either of the hosts {flow.source_node.name} or {flow.destination_node.name} "
                    f"is multi-homed and multi-addressed"
                )
            else:
                self.__log_failure(
                    f"Neither of the hosts {flow.source_node.name} or {flow.destination_node.name} "
                    f"is multi-homed and multi-addressed"
                )

            # Source Node must have atleast 1 interface with
            # MPTCP enabled subflow or fullmesh endpoints
            if any(
                list(
                    map(
                        lambda interface: interface.get_mptcp_endpoints() is not None
                        and any(
                            list(
                                map(
                                    lambda y: "subflow" in y or "fullmesh" in y,
                                    interface.get_mptcp_endpoints().values(),
                                )
                            )
                        ),
                        flow.source_node.interfaces,
                    )
                )
            ):
                self.__log_success(
                    f"Atleast 1 interface on Source Node {flow.source_node.name} "
                    f"has a subflow/fullmesh MPTCP endpoint"
                )
            else:
                self.__log_neutral(
                    f"No interface on Source Node {flow.source_node.name} "
                    f"has a subflow/fullmesh MPTCP endpoint"
                )

            # Destination Node must have atleast 1 interface with
            # MPTCP enabled signal endpoint
            if any(
                list(
                    map(
                        lambda interface: interface.get_mptcp_endpoints() is not None
                        and any(
                            map(
                                lambda y: "signal" in y,
                                interface.get_mptcp_endpoints().values(),
                            )
                        ),
                        flow.destination_node.interfaces,
                    )
                )
            ):
                self.__log_success(
                    f"Atleast 1 interface on Destination Node {flow.destination_node.name} "
                    f"has a signal MPTCP endpoint"
                )
            else:
                self.__log_neutral(
                    f"No interface on Destination Node {flow.destination_node.name} "
                    f"has a signal MPTCP endpoint"
                )
            print(Style.RESET_ALL)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2023 NITK Surathkal

"""
Runs commands to setup MPEG-DASH experiment and collect data
"""

from functools import partial
import logging
from enum import Enum
from threading import Lock

from nest.topology_map import TopologyMap
from ..results import MpegDashResults
from .runnerbase import Runner
from ...engine.mpeg_dash import run_mpeg_dash_client, run_mpeg_dash_http_server
from ..pack import Pack

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-few-public-methods
class MpegDashLogExtractor:
    """
    Contains information related to a single log entry.
    """

    class LogType(Enum):
        """
        An enum depicting the type of the log.
        """

        NONE = "None"
        AUDIO = "Audio"
        VIDEO = "Video"

    def __init__(self, words_in_line):
        self.log_type = MpegDashLogExtractor.LogType.NONE
        self.bitrate_switch = 0
        self.time = 0
        self.segment_size = 0
        self.throughput = 0
        self.bitrate_level = 0
        self.bitrate = 0
        self.buffer = 0
        self.__extract_info(words_in_line)

    def __extract_info(self, words_in_line):
        for i in range(0, len(words_in_line)):
            if words_in_line[i] == "AS#1":
                # Assuming AS1 is Adaptation set for audio
                self.log_type = MpegDashLogExtractor.LogType.AUDIO
            elif words_in_line[i] == "AS#2":
                # Assuming AS2 is Adaptation set of videos
                self.log_type = MpegDashLogExtractor.LogType.VIDEO
            elif words_in_line[i] == "changed" and words_in_line[i + 1] == "quality":
                self.bitrate_switch = 1
                if self.log_type in (
                    MpegDashLogExtractor.LogType.VIDEO,
                    MpegDashLogExtractor.LogType.AUDIO,
                ):
                    # Setting Bitrate switching parameters
                    self.bitrate_level = int(words_in_line[6])
                break
            elif words_in_line[i] == "got":
                self.time = float(words_in_line[8])
                self.throughput = float(words_in_line[11])
                self.segment_size = float(words_in_line[15])
                self.bitrate = float(words_in_line[19])
                self.bitrate_level = int(
                    words_in_line[21][: len(words_in_line[21]) - 1]
                )
                self.buffer = int(words_in_line[25])
                break

    def __str__(self):
        log_data = []
        log_data.append(("Log Type", self.log_type))
        log_data.append(("Bitrate-switch", self.bitrate_switch))
        log_data.append(("Time", self.time))
        log_data.append(("Segment size", self.segment_size))
        log_data.append(("Throughput", self.throughput))
        log_data.append(("Bitrate-level", self.bitrate_level))
        log_data.append(("Bitrate", self.bitrate))
        log_data.append(("Buffer", self.buffer))
        return " " + str(log_data)


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-few-public-methods
class MpegDashRunner(Runner):
    """
    Runs MPEG-DASH client and server.

    """

    lock = Lock()

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        ns_id,
        destination_ip,
        dst_ns,
        port,
        encoded_chunks_path,
        duration,
        player,
        enable_audio_playback,
        additional_player_options,
    ):
        """
        Constructor to initialize the runner

        Parameters
        ----------
        ns_id : str
            network namespace of the client
        destination_ip : str
            the ip of server to which it has to connect
        dst_ns : str
            network namespace of the server
        port: int
            the port number of the server at which it is running
        encoded_chunks_path : Path
            The path where the encoded chunks are present.
        duration : int
            Number of seconds for which experiment has to be run
        player: str
            The media player to be used.
        enable_audio_playback: bool
            Enable/disable audio playback
        additional_player_options : string
            User specified options for the video player
        """
        self.port = port
        self.encoded_chunks_path = encoded_chunks_path
        self.duration = duration
        self.player = player
        self.enable_audio_playback = enable_audio_playback
        self.additional_player_options = additional_player_options
        super().__init__(
            ns_id,
            start_time=None,
            run_time=None,
            destination_ip=destination_ip,
            dst_ns=dst_ns,
        )
        self.audio_logs = []
        self.video_logs = []
        self.audio_info = {
            "bitrate_switches": 0,
            "average_bitrate": 0,
            "average_throughput": 0,
            "average_buffer": 0,
            "average_rtt": 0,
        }
        self.video_info = {
            "bitrate_switches": 0,
            "average_bitrate": 0,
            "average_throughput": 0,
            "average_buffer": 0,
            "average_rtt": 0,
        }

    @staticmethod
    def run_server(ns_id, port, encoded_chunks_path):
        """
        Run MPEG-DASH server in `ns_id`

        Parameters
        ----------
        ns_id : str
            namespace to run MPEG-DASH server on
        port : int
            port to run MPEG-DASH server on
        encoded_chunks_path : Path
            The path where the encoded chunks are present.
        """

        # Run mpeg-dash server
        return_code = run_mpeg_dash_http_server(ns_id, port, encoded_chunks_path)
        if return_code != 0:
            ns_name = TopologyMap.get_node(ns_id).name
            logger.error("Error running MPEG-DASH server at %s.", ns_name)

    # pylint: disable=too-many-statements, too-many-branches
    def run(self):
        """
        Calls engine method to run MPEG-DASH client
        """

        # Run MPEG-DASH client
        super().run(
            partial(
                run_mpeg_dash_client,
                self.ns_id,
                self.destination_address.get_addr(with_subnet=False),
                self.port,
                self.duration,
                player=self.player,
                enable_audio_playback=self.enable_audio_playback,
                additional_player_options=self.additional_player_options,
            ),
            error_string_prefix="Running MPEG-DASH",
        )

    # pylint: disable=too-many-locals
    def parse(self):
        """
        Parse MPEG-DASH output from self.err
        """
        if self.player == "vlc":
            return

        # Rewind to start of the err file and read output
        self.err.seek(0)

        # The dump file name is of the form: 'mpeg_dash_gpac_{client_ns}_to_{server_ns}_
        # ({server_ip}:{server_port}).log' The raw GPAC logs are dumped to this file
        # incase the user wants to parse the logs manually.
        src_ns_name = TopologyMap.get_node(self.ns_id).name
        dst_ns_name = TopologyMap.get_node(self.dst_ns).name

        server_ip_port = (
            self.destination_address.get_addr(with_subnet=False) + ":" + str(self.port)
        )

        dump_filename = "mpeg_dash_gpac_"
        dump_filename += f"{src_ns_name}_to_"
        dump_filename += f"{dst_ns_name}_({server_ip_port}).log"
        Pack.dump_file(dump_filename, self.err.read().decode())

        results_json = {server_ip_port: {}}

        self.err.seek(0)
        file_lines = self.err.readlines()

        lines = []
        for file_line in file_lines:
            lines.extend(file_line.splitlines())

        for line in lines:
            words = list(line.decode().split())
            if (
                len(words) > 1
                and words[0] == "[DASH]"
                and (words[1] == "AS#1" or words[1] == "AS#2")
                and (words[4] != "done")
            ):
                log = MpegDashLogExtractor(words)
                if log.log_type == MpegDashLogExtractor.LogType.AUDIO:
                    self.audio_logs.append(log)
                if log.log_type == MpegDashLogExtractor.LogType.VIDEO:
                    self.video_logs.append(log)

        video_stats_list = [{"server_node": TopologyMap.get_node(self.dst_ns).name}]
        audio_stats_list = [
            {
                "server_node": TopologyMap.get_node(self.dst_ns).name,
                "audio_enabled": False,
            }
        ]
        results_json[server_ip_port]["audio"] = audio_stats_list

        # Obtaining statistics from video logs
        counter = 0
        for log in self.video_logs:
            if log.bitrate_switch:
                self.video_info["bitrate_switches"] += 1
            else:
                counter += 1
                self.video_info["average_bitrate"] += log.bitrate
                self.video_info["average_throughput"] += log.throughput
                self.video_info["average_buffer"] += log.buffer
                self.video_info["average_rtt"] += log.time

        self.video_info["average_bitrate"] /= max(1, counter)
        self.video_info["average_throughput"] /= max(1, counter)
        self.video_info["average_buffer"] /= max(1, counter)
        self.video_info["average_rtt"] /= max(1, counter)

        # Formatting video statistics into JSON
        chunk_number = 0
        for log in self.video_logs:
            if not log.bitrate_switch:
                video_stats_list.append(
                    {
                        "chunk_number": chunk_number,
                        "bitrate": log.bitrate,
                        "bitrate_level": log.bitrate_level,
                        "throughput": log.throughput,
                        "buffer": log.buffer,
                        "rtt": log.time,
                    }
                )
                chunk_number += 1

        video_stats_summary = {
            "number_of_bitrate_switches": self.video_info["bitrate_switches"],
            "average_bitrate": self.video_info["average_bitrate"],
            "average_throughput": self.video_info["average_throughput"],
            "average_buffer": self.video_info["average_buffer"],
            "average_rtt": self.video_info["average_rtt"],
        }

        log_string = "### Video Stream Information ### \n"
        log_string += f"\t ### Client {src_ns_name} to "
        log_string += f"Server {dst_ns_name} "
        log_string += f"({server_ip_port}) ###\n"
        log_string += "\t Number of bitrate switches: "
        log_string += f"{video_stats_summary['number_of_bitrate_switches']} \n"
        log_string += "\t Average Bitrate: "
        log_string += f"{video_stats_summary['average_bitrate']:.2f} Kbps \n"
        log_string += "\t Average Throughput: "
        log_string += f"{video_stats_summary['average_throughput']:.2f} Kbps \n"
        log_string += "\t Average Buffer: "
        log_string += f"{video_stats_summary['average_buffer']:.2f} ms \n"
        log_string += "\t Average RTT: "
        log_string += f"{video_stats_summary['average_rtt']:.2f} s \n\n"

        results_json[server_ip_port]["video"] = video_stats_list
        results_json[server_ip_port]["video_summary"] = video_stats_summary

        if self.enable_audio_playback is True:
            # Obtaining statistics from audio logs
            counter = 0
            for log in self.audio_logs:
                if log.bitrate_switch:
                    self.audio_info["bitrate_switches"] += 1
                else:
                    counter += 1
                    self.audio_info["average_bitrate"] += log.bitrate
                    self.audio_info["average_throughput"] += log.throughput
                    self.audio_info["average_buffer"] += log.buffer
                    self.audio_info["average_rtt"] += log.time

            self.audio_info["average_bitrate"] /= max(1, counter)
            self.audio_info["average_throughput"] /= max(1, counter)
            self.audio_info["average_buffer"] /= max(1, counter)
            self.audio_info["average_rtt"] /= max(1, counter)

            # Formatting audio statistics into JSON
            chunk_number = 0
            for log in self.audio_logs:
                if not log.bitrate_switch:
                    audio_stats_list.append(
                        {
                            "chunk_number": chunk_number,
                            "bitrate": log.bitrate,
                            "bitrate_level": log.bitrate_level,
                            "throughput": log.throughput,
                            "buffer": log.buffer,
                            "rtt": log.time,
                        }
                    )
                    chunk_number += 1

            audio_stats_summary = {
                "number_of_bitrate_switches": self.audio_info["bitrate_switches"],
                "average_bitrate": self.audio_info["average_bitrate"],
                "average_throughput": self.audio_info["average_throughput"],
                "average_buffer": self.audio_info["average_buffer"],
                "average_rtt": self.audio_info["average_rtt"],
            }

            log_string += "\t ### Audio Stream Information ### \n"
            log_string += f"\t ### Client {src_ns_name} to "
            log_string += f"Server {dst_ns_name} "
            log_string += f"({server_ip_port}) ###\n"
            log_string += "\t Number of bitrate switches: "
            log_string += f"{audio_stats_summary['number_of_bitrate_switches']} \n"
            log_string += "\t Average Bitrate: "
            log_string += f"{audio_stats_summary['average_bitrate']:.2f} Kbps \n"
            log_string += "\t Average Throughput: "
            log_string += f"{audio_stats_summary['average_throughput']:.2f} Kbps \n"
            log_string += "\t Average Buffer: "
            log_string += f"{audio_stats_summary['average_buffer']:.2f} ms \n"
            log_string += "\t Average RTT: "
            log_string += f"{audio_stats_summary['average_rtt']:.2f} s \n"

            audio_stats_list[0]["audio_enabled"] = True

            results_json[server_ip_port]["audio"] = audio_stats_list
            results_json[server_ip_port]["audio_summary"] = audio_stats_summary

        with MpegDashRunner.lock:
            logger.info(log_string)
        MpegDashResults.add_result(self.ns_id, results_json)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""IP route commands"""

import os
import json
import inspect
from pathlib import Path
from datetime import datetime
from .exec import exec_subprocess


def add_route(host_name, dest_ip, next_hop_ip, via_int):
    """
    Adds a route in routing table of host.

    Parameters
    ----------
    host_name : str
        name of the host namespace
    dest_ip : str
        the destination ip for the route
    next_hop_ip : str
        IP of the very next interface
    via_int : str
        the corresponding interface in the host
    """
    exec_subprocess(
        f"ip netns exec {host_name} ip route add {dest_ip}"
        f" via {next_hop_ip} dev {via_int}"
    )


def get_route(host_name, *args):
    """
    Get the routes from the routing table of a Node.

    Parameters
    ----------
    host_name : str
        name of the host namespace
    args : Interface
        Interface object of which route information want
    """
    data = exec_subprocess(f"ip netns exec {host_name} ip -j route", False, True)
    date = datetime.now().strftime("%d-%m-%y_%I:%M:%S_%p")
    data = json.loads(data)
    path = Path(inspect.stack()[2][1]).parent
    file_name = Path(inspect.stack()[2][1]).stem
    json_string = json.dumps(data, indent=4)
    complete_name = os.path.join(path, file_name + "_" + date)

    # pylint: disable=bare-except,consider-using-f-string
    if not args:
        try:
            os.mkdir(f"{file_name}_{date}")
        except:
            pass
        with open(
            os.path.join(f"{complete_name}", "All_route_of_" + f"{host_name}.json"), "a"
        ) as outfile:
            outfile.write("{}\n".format(json_string))

    for temp in args:
        try:
            temp.get_address()
        except:
            print(
                " Argument "
                + f" {temp}"
                + "is not in required form, Argument should be Interface"
            )
            continue
        try:
            os.mkdir(f"{complete_name}")
        except:
            pass
        if temp.node_id is not f"{host_name}":
            print(
                "Interface "
                + f"{temp.name}"
                + " is not belong to host "
                + f"{host_name}"
            )
            continue

        for i in data:
            if "prefsrc" in i.keys() and i[
                "prefsrc"
            ].strip() == temp.get_address().get_addr(False):
                with open(
                    os.path.join(
                        f"{complete_name}",
                        "route_info_of_interface_" + f"{temp.name}.json",
                    ),
                    "a",
                ) as outfile:
                    outfile.write("{}\n".format(json.dumps(i, indent=4)))



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""Test thread-safety of the TopologyMap"""

import unittest
from threading import Thread
from nest.topology import connect, Node
from nest.topology_map import TopologyMap

# pylint: disable=missing-docstring
# pylint: disable=too-many-instance-attributes
# pylint: disable=attribute-defined-outside-init


class TestThreadSafety(unittest.TestCase):
    def _add_node(self, node_id):
        # Work done by a thread
        node = Node("n" + str(node_id))
        self.nodes.append(node)

    def _add_and_connect(self, id1, id2):
        # Work done by a thread
        node_1 = Node("n" + str(id1))
        node_2 = Node("n" + str(id2))
        self.nodes.extend([node_1, node_2])

        (eth0, eth1) = connect(node_1, node_2)
        self.interfaces.extend([eth0, eth1])

    def test_thread_safety_of_nodes(self):
        # set up
        num_of_threads = 10

        self.threads = []
        self.nodes = []

        # Create the threads
        for i in range(0, num_of_threads):
            a_thread = Thread(target=self._add_node, args=(i,))
            self.threads.append(a_thread)

        # Start the threads
        for thread in self.threads:
            thread.start()

        # Join the threads
        for thread in self.threads:
            thread.join()

        # Validate
        for node in self.nodes:
            self.assertEqual(TopologyMap.get_node(node.id).name, node.name)

        TopologyMap.delete_all_mapping()

    def test_thread_safety_of_all(self):
        # set up
        num_of_threads = 10

        self.threads = []
        self.nodes = []
        self.interfaces = []
        if num_of_threads % 2 == 1:
            num_of_threads = num_of_threads + 1

        # Create the threads
        for i in range(0, num_of_threads, 2):
            a_thread = Thread(
                target=self._add_and_connect,
                args=(
                    i,
                    i + 1,
                ),
            )
            self.threads.append(a_thread)

        # Start the threads
        for thread in self.threads:
            thread.start()

        # Join the threads
        for thread in self.threads:
            thread.join()

        # Validate
        for node in self.nodes:
            self.assertEqual(TopologyMap.get_node(node.id).name, node.name)

        for interface in self.interfaces:
            self.assertEqual(
                TopologyMap.get_device(interface.node_id, interface.id).name,
                interface.name,
            )

        TopologyMap.delete_all_mapping()


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

""" Runs iperf command to setup UDP flows """

import os
import tempfile
from time import sleep
from functools import partial
import json
import logging
import copy

from nest.experiment.interrupts import handle_keyboard_interrupt
from nest.experiment.pack import Pack
from ..results import Iperf3Results, Iperf3ServerResults
from .runnerbase import Runner
from ...engine.iperf3 import run_iperf_server, run_iperf_client

logger = logging.getLogger(__name__)


class Iperf3Runner(Runner):
    """
    Runs iperf client and server.

    Attributes
    ----------
    ns_id : str
        network namespace to run iperf from
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        ns_id,
        destination_ip,
        bandwidth,
        n_flows,
        start_time,
        run_time,
        dst_ns,
        protocol,
        is_mptcp: bool = False,
        **kwargs,
    ):
        """
        Constructor to initialize the runner

        Parameters
        ----------
        ns_id : str
            network namespace to run iperf3 from
        destination : str
            the ip of server to which it has to connect
        bandwidth : int
            target bandwidth of the UDP flow in mbits
        n_flows : int
            number of parallel flows
        start_time : num
            start time of the flow
        run_time : num
            test duration
        dst_ns: str
            network namespace to run iperf3 server from
        protocol: str
            transport layer protocol, either "tcp" or "udp"
        is_mptcp : bool
            boolean to determine if connection is MPTCP enabled
        """
        self.bandwidth = bandwidth
        self.n_flows = n_flows
        self.protocol = protocol
        self.options = copy.deepcopy(kwargs)
        self.is_mptcp = is_mptcp
        super().__init__(ns_id, start_time, run_time, destination_ip, dst_ns)

    def setup_iperf3_client(self, options):
        """
        Setup iperf3 client with Default value

        Parameters
        ----------
        options : dict
            iperf3 client options configured by user
        """

        # TOOD: Only iperf3 flags should be added here (without any default values)
        client_option_flags = {
            "verbose": "-V",  # more detailed output as a log file
            "format": "-f ",  # format to report: Kbits, Mbits, Gbits, Tbits
            "logfile": "--logfile ",  # send output to a log file
            "forceflush": "--forceflush",  # force flushing output at every interval
            "port_no": "-p ",  # # server port to connect to
            "cport": "--cport ",  # bind to a specific client port
            "interval": "-i ",  # Generate interim results every INTERVAL seconds
            "n_flows": "-P 1",  # Number of parallel flows (NOTE: Default 1)
            "testlen": "-t 10",  # Length of test (NOTE: Default 10s)
            "cong_algo": "-C ",  # Congestion algorithm
            "bind": "-B ",  # Bind to a specific interface
        }

        client_options = {"json": "--json"}
        # if iperf3 contains any of the below options then output will be a .log file
        loglist = [
            "format",
            "timestamps",
            "verbose",
            "forceflush",
            "logfile",
        ]
        for listitem in loglist:
            if listitem in options:
                client_options.pop("json")
                if "logfile" not in options:
                    options.update({"logfile": "Iperf3_Client_Output"})
                options.update({"logfile": options["logfile"] + ".log"})
                break
        for option in options:
            if option in client_option_flags:
                option_value = client_option_flags[option]
                if not isinstance(options[option], bool):
                    option_value += str(options[option])
                client_options.update({option: option_value})

        if self.protocol == "udp":
            client_options.update({"protocol": "--udp"})

        if "timestamps" in options:
            client_options.update(
                {"timestamps": f"--timestamps=\"{options['timestamps']} \""}
            )

        self.options = client_options

    def run(self):
        """
        calls engine method to run iperf client
        """
        # set user defined iperf3 client options
        iperf3_options = self.options

        # Set run time
        iperf3_options["testlen"] = f"-t {self.run_time}"

        # Set number of parallel flows
        iperf3_options["n_flows"] = f"-P {self.n_flows}"

        # Set target bitrate
        if "protocol" in iperf3_options and iperf3_options["protocol"] == "--udp":
            iperf3_options["bitrate"] = (
                f"-b {self.bandwidth}" if self.bandwidth else "-b 1mbit"
            )

        # Convert iperf3_options dict to string
        iperf3_options_list = list(iperf3_options.values())
        iperf3_options_string = " ".join(iperf3_options_list)

        if self.start_time != 0:
            sleep(self.start_time)

        waiting_time = self.run_time
        while True:
            super().run(
                partial(
                    run_iperf_client,
                    self.ns_id,
                    iperf3_options_string,
                    self.destination_address.get_addr(with_subnet=False),
                    self.destination_address.is_ipv6(),
                    self.is_mptcp,
                ),
                error_string_prefix="Running iperf3 Client",
            )
            self.out.seek(0)
            out_str = self.out.read().decode()
            parse = json.loads(out_str)
            if not "error" in parse:
                break
            if waiting_time < 2:
                dst_address = self.destination_address.get_addr(with_subnet=False)
                logger.error("unable to connect server on destination %s", dst_address)
                break
            logger.info("%s\nRetrying to connect ...\n", parse["error"])
            self.out.close()

            # pylint: disable=consider-using-with
            self.out = tempfile.TemporaryFile()
            waiting_time -= 1
            sleep(1)

    # pylint: disable=too-many-locals
    @handle_keyboard_interrupt
    def parse(self):
        """
        Parse iperf3 output from self.out
        """
        if "logfile" in self.options:
            logfile_name = self.options["logfile"].replace("--logfile ", "")
            logfile_path = os.path.join(os.getcwd(), logfile_name)
            if os.path.exists(logfile_path):
                Pack.move_files(logfile_path)
            return
        self.out.seek(0)  # rewind to start of the temp file
        raw_stats = self.out.read().decode()

        # Iperf3 already gives output in JSON format, so we parse
        # into a dictionary
        parsed_stats = json.loads(raw_stats)
        if "error" in parsed_stats:
            return

        # Extract from useful connection info from parsed data
        # Useful for diffrentiating between flows
        start_timestamp = parsed_stats["start"]["timestamp"]["timesecs"]
        connection_info = {}

        for conn in parsed_stats["start"]["connected"]:
            connection_info[conn["socket"]] = {
                "local_host": conn["local_host"],
                "local_port": conn["local_port"],
                "remote_host": conn["remote_host"],
                "remote_port": conn["remote_port"],
            }

        # Convert iperf3 JSON data into our own JSON format (similar
        # to netperf), so that plotter code can be reused

        stats_dict_list = {}
        for socket, socket_info in connection_info.items():
            local_port = socket_info["local_port"]
            stats_dict_list[local_port] = [self.get_meta_item()]

        if len(stats_dict_list) > 1:
            stats_dict_list["sum"] = [self.get_meta_item()]

        for interval in parsed_stats["intervals"]:
            for stream in interval["streams"]:
                socket = stream["socket"]
                local_port = connection_info[socket]["local_port"]
                stats_dict_list[local_port].append(
                    self._extract_from_iperf3_stream(start_timestamp, stream)
                )

            if len(stats_dict_list) > 1:
                sum_stream = interval["sum"]
                stats_dict_list["sum"].append(
                    self._extract_from_iperf3_stream(start_timestamp, sum_stream)
                )

        destination_ip = self.destination_address.get_addr(with_subnet=False)
        Iperf3Results.add_result(self.ns_id, {destination_ip: stats_dict_list})

    def _extract_from_iperf3_stream(self, start_timestamp, stream):
        """
        Convert information in iperf3 stream into required
        dictionary format
        """
        if self.protocol == "udp":
            return {
                "timestamp": str(stream["start"] + start_timestamp),
                "sending_rate": str(stream["bits_per_second"] / 1e6),
                "duration": str(stream["seconds"]),
                "bytes": str(stream["bytes"]),
                "packets": str(stream["packets"]),
            }
        return {
            "timestamp": str(stream["start"] + start_timestamp),
            "sending_rate": str(stream["bits_per_second"] / 1e6),
            "duration": str(stream["seconds"]),
            "bytes": str(stream["bytes"]),
        }


class Iperf3ServerRunner(Runner):
    """
    Runs iperf3 server. Currently being used
    for UDP flows only

    Attributes
    ----------
    ns_id : str
        network namespace to run iperf3 server
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, ns_id, run_time, protocol, is_mptcp):
        """
                Constructor to initialize the runner

        Parameters
        ----------
        ns_id : str
            network namespace to run iperf3 server
        run_time : num
            test duration
        protocol: str
            transport layer protocol, either "tcp" or "udp"
        is_mptcp : bool
            boolean to determine if connection is MPTCP enabled
        """
        self.ns_id = ns_id
        self.protocol = protocol
        self.server_options = {}
        self.is_mptcp = is_mptcp
        super().__init__(ns_id, 0, run_time)

    def setup_iperf3_server(self, options):
        """
        Setup iperf3 server options with Default value

        Parameters
        ----------
        options : dict
            iperf3 server options configured by user
        """

        default_server_options = {
            "one_off": "-1",  # handle one client connection then exit
            "port_no": "-p ",  # server port to listen on
            "s_verbose": "-V",  # more detailed output as a log file
            "s_interval": "-i ",  # Generate interim results every INTERVAL seconds
            "s_format": "-f ",  # format to report: Kbits, Mbits, Gbits, Tbits
            "s_logfile": "--logfile ",  # send output to a log file
            "s_forceflush": "--forceflush",  # force flushing output at every interval
            "bitrate": "--server-bitrate-limit ",  # server's total bit rate limit
            "bind": "-B ",  # Bind to a specific interface
        }

        # Default Server output is a .json file
        server_options = {"json": "--json"}

        # if iperf3 contains any of the below options then output will be a .log file
        loglist = [
            "s_format",
            "s_timestamps",
            "s_verbose",
            "s_forceflush",
            "s_logfile",
            "daemon",
        ]
        for listitem in loglist:
            if listitem in options:
                server_options.pop("json")
                if "s_logfile" not in options:
                    options.update({"s_logfile": "Iperf3_Server_Output"})
                options.update({"s_logfile": options["s_logfile"] + ".log"})
                break

        for option in options:
            if option in default_server_options:
                option_value = default_server_options[option]
                if not isinstance(options[option], bool):
                    option_value += str(options[option])
                server_options.update({option: option_value})
        if "s_timestamps" in options:
            server_options.update(
                {"s_timestamps": f"--timestamps=\"{options['s_timestamps']} \""}
            )

        self.server_options = server_options

    # pylint: disable = invalid-name
    def run(self):
        """
        calls engine method to run iperf3 server
        """
        iperf3_options = self.server_options

        # Convert iperf3_options dict to string
        iperf3_options_list = list(iperf3_options.values())
        iperf3_options_string = " ".join(iperf3_options_list)

        super().run(
            partial(run_iperf_server, self.ns_id, iperf3_options_string, self.is_mptcp),
            error_string_prefix="Running iperf3 server",
        )

    # pylint: disable = too-many-locals
    def parse(self):
        """
        Parse iperf3 output from self.out
        """
        if "s_logfile" in self.server_options:
            logfile_name = self.server_options.get("s_logfile").replace(
                "--logfile ", ""
            )
            logfile_path = os.path.join(os.getcwd(), logfile_name)
            if os.path.exists(logfile_path):
                Pack.move_files(logfile_path)
            return

        self.out.seek(0)  # rewind to start of the temp file
        raw_stats = self.out.read()
        if not raw_stats.decode():
            return

        # Iperf3 already gives output in JSON format, so we parse
        # into a dictionary
        parsed_stats = json.loads(raw_stats.decode())

        # Extract from useful connection info from parsed data
        # Useful for diffrentiating between flows
        start_timestamp = parsed_stats["start"]["timestamp"]["timesecs"]
        connection_info = {}

        for conn in parsed_stats["start"]["connected"]:
            connection_info[conn["socket"]] = {
                "local_host": conn["local_host"],
                "local_port": conn["local_port"],
                "remote_host": conn["remote_host"],
                "remote_port": conn["remote_port"],
            }

        # Convert iperf3 JSON data into our own JSON format (similar
        # to netperf), so that plotter code can be reused

        stats_dict_list = {}
        for socket, socket_info in connection_info.items():
            local_port = socket_info["local_port"]
            destination_ip = socket_info["remote_host"]
            stats_dict_list[local_port] = [self.get_meta_item()]

        if len(stats_dict_list) > 1:
            stats_dict_list["sum"] = [self.get_meta_item()]

        for interval in parsed_stats["intervals"]:
            for stream in interval["streams"]:
                socket = stream["socket"]
                local_port = connection_info[socket]["local_port"]
                stats_dict_list[local_port].append(
                    self._extract_from_iperf3_stream(start_timestamp, stream)
                )

            if len(stats_dict_list) > 1:
                sum_stream = interval["sum"]
                stats_dict_list["sum"].append(
                    self._extract_from_iperf3_stream(start_timestamp, sum_stream)
                )
        Iperf3ServerResults.add_result(self.ns_id, {destination_ip: stats_dict_list})

    def _extract_from_iperf3_stream(self, start_timestamp, stream):
        """
        Convert information in iperf3 stream into required
        dictionary format
        """
        if self.protocol == "udp":
            return {
                "timestamp": str(stream["start"] + start_timestamp),
                "receiving_rate": str(stream["bits_per_second"] / 1e6),
                "duration": str(stream["seconds"]),
                "bytes": str(stream["bytes"]),
                "packets": str(stream["packets"]),
            }
        return {
            "timestamp": str(stream["start"] + start_timestamp),
            "receiving_rate": str(stream["bits_per_second"] / 1e6),
            "duration": str(stream["seconds"]),
            "bytes": str(stream["bytes"]),
        }
